# -*- coding: utf-8; -*-
"""
Copyright (c) 2019 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

"""

import sys
from configparser import ConfigParser
from copy import deepcopy
from os.path import expanduser, join, isfile, dirname
from os.path import splitext

import psutil
from numpy import uint8

from exceptions import ArgumentError
from miscellaneous import Miscellaneous

# Set the current software version.
PSS_Version = "PlanetarySystemStacker 0.9.8"
# PSS_Version = "PlanetarySystemStacker"


class ConfigurationParameters(object):
    def __init__(self):
        self.hidden_parameters_current_dir = None
        self.hidden_parameters_main_window_x0 = None
        self.hidden_parameters_main_window_y0 = None
        self.hidden_parameters_main_window_width = None
        self.hidden_parameters_main_window_height = None
        self.hidden_parameters_main_window_maximized = None
        self.global_parameters_display_quickstart = None
        self.global_parameters_version = None
        self.global_parameters_protocol_level = None
        self.global_parameters_write_protocol_to_file = None
        self.global_parameters_store_protocol_with_result = None
        self.global_parameters_buffering_level = None
        self.global_parameters_maximum_memory_active = None
        self.global_parameters_maximum_memory_amount = None
        self.global_parameters_include_postprocessing = None
        self.global_parameters_image_format = None
        self.global_parameters_parameters_in_filename = None
        self.global_parameters_stack_number_frames = None
        self.global_parameters_stack_percent_frames = None
        self.global_parameters_ap_box_size = None
        self.global_parameters_ap_number = None
        self.frames_gauss_width = None
        self.frames_debayering_default = None
        self.frames_debayering_method = None
        self.frames_normalization = None
        self.frames_normalization_threshold = None
        self.frames_add_selection_dialog = None
        self.align_frames_fast_changing_object = None
        self.align_frames_mode = None
        self.align_frames_automation = None
        self.align_frames_rectangle_scale_factor = None
        self.align_frames_search_width = None
        self.align_frames_average_frame_percent = None
        self.alignment_points_half_box_width = None
        self.alignment_points_search_width = None
        self.alignment_points_structure_threshold = None
        self.alignment_points_brightness_threshold = None
        self.alignment_points_frame_percent = None
        self.alignment_points_frame_number = None
        self.stack_frames_drizzle_factor_string = None

    def set_defaults(self):
        self.hidden_parameters_current_dir = expanduser("~")
        self.hidden_parameters_main_window_x0 = 100
        self.hidden_parameters_main_window_y0 = 100
        self.hidden_parameters_main_window_width = 1200
        self.hidden_parameters_main_window_height = 800
        self.hidden_parameters_main_window_maximized = False
        self.global_parameters_display_quickstart = True
        self.global_parameters_version = PSS_Version
        self.global_parameters_protocol_level = 1
        self.global_parameters_write_protocol_to_file = False
        self.global_parameters_store_protocol_with_result = False
        self.global_parameters_buffering_level = -1
        self.global_parameters_maximum_memory_active = False
        self.global_parameters_maximum_memory_amount = \
            max(1, int(dict(psutil.virtual_memory()._asdict())['available'] / 1e9))
        self.global_parameters_include_postprocessing = True
        self.global_parameters_image_format = "png"
        self.global_parameters_parameters_in_filename = False
        self.global_parameters_stack_number_frames = False
        self.global_parameters_stack_percent_frames = False
        self.global_parameters_ap_box_size = False
        self.global_parameters_ap_number = False
        self.frames_gauss_width = 7
        self.frames_debayering_default = 'Auto detect color'
        self.frames_debayering_method = 'Bilinear'
        self.frames_normalization = True
        self.frames_normalization_threshold = 15
        self.frames_add_selection_dialog = False
        self.align_frames_fast_changing_object = True
        self.align_frames_mode = 'Surface'
        self.align_frames_automation = True
        self.align_frames_rectangle_scale_factor = 3.
        self.align_frames_search_width = 34
        self.align_frames_average_frame_percent = 5
        self.alignment_points_search_width = 14
        self.alignment_points_frame_percent = 10
        self.alignment_points_frame_number = -1
        self.stack_frames_drizzle_factor_string = "Off"
        self.set_defaults_ap_editing()

    def set_defaults_ap_editing(self):
        self.alignment_points_half_box_width = 24
        self.alignment_points_structure_threshold = 0.04
        self.alignment_points_brightness_threshold = 10

    def copy_from_config_object(self, configuration_object):
        self.hidden_parameters_current_dir = configuration_object.hidden_parameters_current_dir
        self.hidden_parameters_main_window_x0 = \
            configuration_object.hidden_parameters_main_window_x0
        self.hidden_parameters_main_window_y0 = \
            configuration_object.hidden_parameters_main_window_y0
        self.hidden_parameters_main_window_width = \
            configuration_object.hidden_parameters_main_window_width
        self.hidden_parameters_main_window_height = \
            configuration_object.hidden_parameters_main_window_height
        self.hidden_parameters_main_window_maximized = \
            configuration_object.hidden_parameters_main_window_maximized
        self.global_parameters_display_quickstart = \
            configuration_object.global_parameters_display_quickstart
        self.global_parameters_version = configuration_object.global_parameters_version
        self.global_parameters_protocol_level = \
            configuration_object.global_parameters_protocol_level
        self.global_parameters_write_protocol_to_file = \
            configuration_object.global_parameters_write_protocol_to_file
        self.global_parameters_store_protocol_with_result = \
            configuration_object.global_parameters_store_protocol_with_result
        self.global_parameters_buffering_level = \
            configuration_object.global_parameters_buffering_level
        self.global_parameters_maximum_memory_active = \
            configuration_object.global_parameters_maximum_memory_active
        self.global_parameters_maximum_memory_amount = \
            configuration_object.global_parameters_maximum_memory_amount
        self.global_parameters_include_postprocessing = \
            configuration_object.global_parameters_include_postprocessing
        self.global_parameters_image_format = \
            configuration_object.global_parameters_image_format
        self.global_parameters_parameters_in_filename = \
            configuration_object.global_parameters_parameters_in_filename
        self.global_parameters_stack_number_frames = \
            configuration_object.global_parameters_stack_number_frames
        self.global_parameters_stack_percent_frames = \
            configuration_object.global_parameters_stack_percent_frames
        self.global_parameters_ap_box_size = \
            configuration_object.global_parameters_ap_box_size
        self.global_parameters_ap_number = \
            configuration_object.global_parameters_ap_number
        self.frames_gauss_width = configuration_object.frames_gauss_width
        self.frames_debayering_default = configuration_object.frames_debayering_default
        self.frames_debayering_method = configuration_object.frames_debayering_method
        self.frames_normalization = configuration_object.frames_normalization
        self.frames_normalization_threshold = configuration_object.frames_normalization_threshold
        self.frames_add_selection_dialog = configuration_object.frames_add_selection_dialog
        self.align_frames_fast_changing_object = \
            configuration_object.align_frames_fast_changing_object
        self.align_frames_mode = configuration_object.align_frames_mode
        self.align_frames_automation = configuration_object.align_frames_automation
        self.align_frames_rectangle_scale_factor = \
            configuration_object.align_frames_rectangle_scale_factor
        self.align_frames_search_width = configuration_object.align_frames_search_width
        self.align_frames_average_frame_percent = \
            configuration_object.align_frames_average_frame_percent
        self.alignment_points_half_box_width = configuration_object.alignment_points_half_box_width
        self.alignment_points_search_width = configuration_object.alignment_points_search_width
        self.alignment_points_structure_threshold = \
            configuration_object.alignment_points_structure_threshold
        self.alignment_points_brightness_threshold = \
            configuration_object.alignment_points_brightness_threshold
        self.alignment_points_frame_percent = configuration_object.alignment_points_frame_percent
        self.alignment_points_frame_number = configuration_object.alignment_points_frame_number
        self.stack_frames_drizzle_factor_string = \
            configuration_object.stack_frames_drizzle_factor_string


class Configuration(object):
    def __init__(self):
        # self.global_parameters_version = "PlanetarySystemStacker"
        self.global_parameters_version = PSS_Version

        # Set fixed parameters which are hidden from the user. Hidden parameters which are
        # changeable are stored in the configuration object.

        # Look for PSS icon in several places:
        python_dir = dirname(sys.executable)
        icon_locations = ['PSS-Icon-64.png', join('Icons', 'PSS-Icon-64.png'),
                          join(python_dir, "Lib", "site-packages", 'planetary_system_stacker',
                                'Icons', 'PSS-Icon-64.png')]

        self.window_icon = None
        for location in icon_locations:
            if isfile(location):
                self.window_icon = location
                break

        self.frames_mono_channel = 'panchromatic'
        self.frames_color_difference_threshold = 0
        self.frames_bayer_max_noise_diff_green = 2.
        self.frames_bayer_min_distance_from_blue = 99.5

        self.rank_frames_pixel_stride = 2
        self.rank_frames_method = "Laplace"

        self.align_frames_method = "MultiLevelCorrelation"
        self.align_frames_rectangle_black_threshold = 10240
        self.align_frames_rectangle_min_fraction = 0.7
        self.align_frames_rectangle_stride = 2
        self.align_frames_border_width = 10
        self.align_frames_sampling_stride = 2
        self.align_frames_min_stabilization_patch_fraction = 0.2
        self.align_frames_max_stabilization_patch_fraction = 0.7
        self.align_frames_max_search_width = 150
        self.align_frames_best_frames_window_extension = 2

        self.alignment_points_min_half_box_width = 10
        self.alignment_points_contrast_threshold = 0
        self.alignment_points_dim_fraction_threshold = 0.6
        self.alignment_points_rank_method = "Laplace"
        self.alignment_points_rank_pixel_stride = 2
        self.alignment_points_de_warp = True
        self.alignment_points_method = 'MultiLevelCorrelation'
        self.alignment_points_sampling_stride = 2
        self.alignment_points_local_search_subpixel = False
        self.alignment_points_penalty_factor = 0.00025

        self.stack_frames_suffix = "_pss"
        self.stack_frames_background_blend_threshold = 0.2
        self.stack_frames_background_fraction = 0.3
        self.stack_frames_background_patch_size = 100

        self.postproc_suffix = "_gpp"
        self.postproc_max_layers = 10
        self.postproc_bi_range_standard = 13
        self.postproc_max_shift = 5
        self.postproc_max_ram_percentage = 20.
        self.postproc_blinking_period = 1.
        self.postproc_idle_loop_time = 0.2

        # Initialize the ConfigParser object for parameters which the user can change.
        self.config_parser_object = ConfigParser()

        # Create and initialize the central data object for postprocessing.
        self.postproc_data_object = PostprocDataObject(self.postproc_suffix)

    def initialize_configuration(self, read_from_file=True):
        """
        Initialize the configuration with parameters stored during the previous PSS run. This
        initialization is skipped if PSS is executed from the command line. Finally, set derived
        parameters, i.e. parameters which are computed from other parameters.

        :param read_from_file: If True, read parameters from a configuration file.
                               If False, skip this step.
        :return: -
        """

        if read_from_file:
            # The config file for persistent parameter storage is located in the user's home
            # directory, as is the detailed logfile.
            self.home = expanduser("~")
            self.config_filename = join(self.home, ".PlanetarySystemStacker.ini")
            self.protocol_filename = join(self.home, "PlanetarySystemStacker.log")

            # Determine if there is a configuration file from a previous run.
            self.config_file_exists = isfile(self.config_filename)
        else:
            self.config_file_exists = False

        # If an existing config file is found, read it in. Set flag to indicate if parameters were
        # read from file successfully.
        self.configuration_read = False
        if self.config_file_exists:
            try:
                self.config_parser_object = self.read_config()
                self.configuration_read = True
            except:
                self.configuration_read = False

        # The configuration could not be read from a file, or versions did not match. Create a
        # new set of standard parameters.
        if not self.configuration_read:
            configuration_parameters = ConfigurationParameters()
            configuration_parameters.set_defaults()

            # Set current configuration parameters to the new values.
            self.import_from_configuration_parameters(configuration_parameters)

            # Create and initialize the central data object for postprocessing.
            self.postproc_data_object = PostprocDataObject(self.postproc_suffix)

        # Compute parameters which are derived from other parameters.
        self.set_derived_parameters()

        # Mark the configuration as not changed.
        self.configuration_changed = False
        self.go_back_to_activity = None

    def import_from_configuration_parameters(self, configuration_parameters):
        """
        Set all current parameters to the corresponding values of a ConfigurationParameters object.

        :param configuration_parameters: ConfigurationParameters object with new parameter values.
        :return: -
        """

        self.hidden_parameters_current_dir = configuration_parameters.hidden_parameters_current_dir
        self.hidden_parameters_main_window_x0 = \
            configuration_parameters.hidden_parameters_main_window_x0
        self.hidden_parameters_main_window_y0 = \
            configuration_parameters.hidden_parameters_main_window_y0
        self.hidden_parameters_main_window_width = \
            configuration_parameters.hidden_parameters_main_window_width
        self.hidden_parameters_main_window_height = \
            configuration_parameters.hidden_parameters_main_window_height
        self.hidden_parameters_main_window_maximized = \
            configuration_parameters.hidden_parameters_main_window_maximized
        self.global_parameters_display_quickstart = \
            configuration_parameters.global_parameters_display_quickstart
        self.global_parameters_version = configuration_parameters.global_parameters_version
        self.global_parameters_protocol_level = \
            configuration_parameters.global_parameters_protocol_level
        self.global_parameters_write_protocol_to_file = \
            configuration_parameters.global_parameters_write_protocol_to_file
        self.global_parameters_store_protocol_with_result = \
            configuration_parameters.global_parameters_store_protocol_with_result
        self.global_parameters_buffering_level = \
            configuration_parameters.global_parameters_buffering_level
        self.global_parameters_maximum_memory_active = \
            configuration_parameters.global_parameters_maximum_memory_active
        self.global_parameters_maximum_memory_amount = \
            configuration_parameters.global_parameters_maximum_memory_amount
        self.global_parameters_include_postprocessing = \
            configuration_parameters.global_parameters_include_postprocessing
        self.global_parameters_image_format = \
            configuration_parameters.global_parameters_image_format
        self.global_parameters_parameters_in_filename = \
            configuration_parameters.global_parameters_parameters_in_filename
        self.global_parameters_stack_number_frames = \
            configuration_parameters.global_parameters_stack_number_frames
        self.global_parameters_stack_percent_frames = \
            configuration_parameters.global_parameters_stack_percent_frames
        self.global_parameters_ap_box_size = \
            configuration_parameters.global_parameters_ap_box_size
        self.global_parameters_ap_number = \
            configuration_parameters.global_parameters_ap_number
        self.frames_gauss_width = configuration_parameters.frames_gauss_width
        self.frames_debayering_default = configuration_parameters.frames_debayering_default
        self.frames_debayering_method = configuration_parameters.frames_debayering_method
        self.frames_normalization = configuration_parameters.frames_normalization
        self.frames_normalization_threshold = \
            configuration_parameters.frames_normalization_threshold
        self.frames_add_selection_dialog = \
            configuration_parameters.frames_add_selection_dialog
        self.align_frames_fast_changing_object = \
            configuration_parameters.align_frames_fast_changing_object
        self.align_frames_mode = configuration_parameters.align_frames_mode
        self.align_frames_automation = configuration_parameters.align_frames_automation
        self.align_frames_rectangle_scale_factor = \
            configuration_parameters.align_frames_rectangle_scale_factor
        self.align_frames_search_width = configuration_parameters.align_frames_search_width
        self.align_frames_average_frame_percent = \
            configuration_parameters.align_frames_average_frame_percent
        self.alignment_points_half_box_width = \
            configuration_parameters.alignment_points_half_box_width
        self.alignment_points_search_width = configuration_parameters.alignment_points_search_width
        self.alignment_points_structure_threshold = \
            configuration_parameters.alignment_points_structure_threshold
        self.alignment_points_brightness_threshold = \
            configuration_parameters.alignment_points_brightness_threshold
        self.alignment_points_frame_percent = \
            configuration_parameters.alignment_points_frame_percent
        self.alignment_points_frame_number = \
            configuration_parameters.alignment_points_frame_number
        self.stack_frames_drizzle_factor_string = \
            configuration_parameters.stack_frames_drizzle_factor_string

    def export_to_configuration_parameters(self, configuration_parameters):
        """
        Set all values of a ConfigurarionParameters object to the current parameters.

        :param configuration_parameters: ConfigurarionParameters object to be updated.
        :return: -
        """

        configuration_parameters.hidden_parameters_current_dir = self.hidden_parameters_current_dir
        configuration_parameters.hidden_parameters_main_window_x0 = \
            self.hidden_parameters_main_window_x0
        configuration_parameters.hidden_parameters_main_window_y0 = \
            self.hidden_parameters_main_window_y0
        configuration_parameters.hidden_parameters_main_window_width = \
            self.hidden_parameters_main_window_width
        configuration_parameters.hidden_parameters_main_window_height = \
            self.hidden_parameters_main_window_height
        configuration_parameters.hidden_parameters_main_window_maximized = \
            self.hidden_parameters_main_window_maximized
        configuration_parameters.global_parameters_display_quickstart = \
            self.global_parameters_display_quickstart
        configuration_parameters.global_parameters_version = self.global_parameters_version
        configuration_parameters.global_parameters_protocol_level = \
            self.global_parameters_protocol_level
        configuration_parameters.global_parameters_write_protocol_to_file = \
            self.global_parameters_write_protocol_to_file
        configuration_parameters.global_parameters_store_protocol_with_result = \
            self.global_parameters_store_protocol_with_result
        configuration_parameters.global_parameters_include_postprocessing = \
            self.global_parameters_include_postprocessing
        configuration_parameters.global_parameters_image_format = \
            self.global_parameters_image_format
        configuration_parameters.global_parameters_parameters_in_filename = \
            self.global_parameters_parameters_in_filename
        configuration_parameters.global_parameters_stack_number_frames = \
            self.global_parameters_stack_number_frames
        configuration_parameters.global_parameters_stack_percent_frames = \
            self.global_parameters_stack_percent_frames
        configuration_parameters.global_parameters_ap_box_size = \
            self.global_parameters_ap_box_size
        configuration_parameters.global_parameters_ap_number = \
            self.global_parameters_ap_number

        configuration_parameters.frames_gauss_width = self.frames_gauss_width
        configuration_parameters.frames_debayering_default = self.frames_debayering_default
        configuration_parameters.frames_debayering_method = self.frames_debayering_method
        configuration_parameters.frames_normalization = self.frames_normalization
        configuration_parameters.frames_normalization_threshold = self.frames_normalization_threshold
        configuration_parameters.frames_add_selection_dialog = self.frames_add_selection_dialog

        configuration_parameters.align_frames_fast_changing_object = \
            self.align_frames_fast_changing_object
        configuration_parameters.align_frames_mode = self.align_frames_mode
        configuration_parameters.align_frames_automation = self.align_frames_automation
        configuration_parameters.align_frames_rectangle_scale_factor = \
            self.align_frames_rectangle_scale_factor
        configuration_parameters.align_frames_search_width = self.align_frames_search_width
        configuration_parameters.align_frames_average_frame_percent = \
            self.align_frames_average_frame_percent

        configuration_parameters.alignment_points_half_box_width = \
            self.alignment_points_half_box_width
        configuration_parameters.alignment_points_search_width = self.alignment_points_search_width
        configuration_parameters.alignment_points_structure_threshold = \
            self.alignment_points_structure_threshold
        configuration_parameters.alignment_points_brightness_threshold = \
            self.alignment_points_brightness_threshold
        configuration_parameters.alignment_points_frame_percent = \
            self.alignment_points_frame_percent
        configuration_parameters.alignment_points_frame_number = \
            self.alignment_points_frame_number
        configuration_parameters.stack_frames_drizzle_factor_string = \
            self.stack_frames_drizzle_factor_string

    def get_all_parameters_from_configparser(self, conf):
        """
        All parameters which can be modified by the user are stored in the ConfigParser object.
        This way they can be written to a file or read from there. Read parameter values from a
        ConfigParser object and store them with the configuration object.

        :param conf: ConfigParser object
        :return: -
        """

        # Create an object with default parameters. They are inserted if the corresponding value
        # cannot be read from the config parser object.
        default_conf_obj = ConfigurationParameters()
        default_conf_obj.set_defaults()

        # Read the PSS version with which the ini file was created.
        self.global_parameters_version_imported_from = get_from_conf(conf, 'Global parameters',
            'version', ' version number could not be identified')

        self.hidden_parameters_current_dir = get_from_conf(conf, 'Hidden parameters',
            'current directory', default_conf_obj.hidden_parameters_current_dir)
        self.hidden_parameters_main_window_x0 = get_from_conf(conf, 'Hidden parameters',
            'main window x0', default_conf_obj.hidden_parameters_main_window_x0)
        self.hidden_parameters_main_window_y0 = get_from_conf(conf, 'Hidden parameters',
            'main window y0', default_conf_obj.hidden_parameters_main_window_y0)
        self.hidden_parameters_main_window_width = get_from_conf(conf, 'Hidden parameters',
            'main window width', default_conf_obj.hidden_parameters_main_window_width)
        self.hidden_parameters_main_window_height = get_from_conf(conf, 'Hidden parameters',
            'main window height', default_conf_obj.hidden_parameters_main_window_height)
        self.hidden_parameters_main_window_maximized = get_from_conf(conf, 'Hidden parameters',
            'main window maximized', default_conf_obj.hidden_parameters_main_window_maximized)

        self.global_parameters_display_quickstart = get_from_conf(conf, 'Global parameters',
            'display quickstart', default_conf_obj.global_parameters_display_quickstart)
        self.global_parameters_protocol_level = get_from_conf(conf, 'Global parameters',
            'protocol level', default_conf_obj.global_parameters_protocol_level)
        self.global_parameters_write_protocol_to_file = get_from_conf(conf, 'Global parameters',
            'write protocol to file', default_conf_obj.global_parameters_write_protocol_to_file)
        self.global_parameters_store_protocol_with_result = get_from_conf(conf, 'Global parameters',
            'store protocol with result', default_conf_obj.global_parameters_store_protocol_with_result)
        self.global_parameters_buffering_level = get_from_conf(conf, 'Global parameters',
            'buffering level', default_conf_obj.global_parameters_buffering_level)
        self.global_parameters_maximum_memory_active = get_from_conf(conf, 'Global parameters',
            'maximum memory active', default_conf_obj.global_parameters_maximum_memory_active)
        self.global_parameters_maximum_memory_amount = get_from_conf(conf, 'Global parameters',
            'maximum memory amount', default_conf_obj.global_parameters_maximum_memory_amount)
        self.global_parameters_include_postprocessing = get_from_conf(conf, 'Global parameters',
            'include postprocessing', default_conf_obj.global_parameters_include_postprocessing)
        self.global_parameters_image_format = get_from_conf(conf, 'Global parameters',
            'image format', default_conf_obj.global_parameters_image_format)
        self.global_parameters_parameters_in_filename = get_from_conf(conf, 'Global parameters',
            'parameters in filename', default_conf_obj.global_parameters_parameters_in_filename)
        self.global_parameters_stack_number_frames = get_from_conf(conf, 'Global parameters',
            'stack number frames', default_conf_obj.global_parameters_stack_number_frames)
        self.global_parameters_stack_percent_frames = get_from_conf(conf, 'Global parameters',
            'stack percent frames', default_conf_obj.global_parameters_stack_percent_frames)
        self.global_parameters_ap_box_size = get_from_conf(conf, 'Global parameters',
            'ap box size', default_conf_obj.global_parameters_ap_box_size)
        self.global_parameters_ap_number = get_from_conf(conf, 'Global parameters',
            'ap number', default_conf_obj.global_parameters_ap_number)

        self.frames_gauss_width = get_from_conf(conf, 'Frames',
            'gauss width', default_conf_obj.frames_gauss_width)
        self.frames_debayering_default = get_from_conf(conf, 'Frames',
            'debayering default', default_conf_obj.frames_debayering_default)
        self.frames_debayering_method = get_from_conf(conf, 'Frames',
            'debayering method', default_conf_obj.frames_debayering_method)
        self.frames_normalization = get_from_conf(conf, 'Frames',
            'normalization', default_conf_obj.frames_normalization)
        self.frames_normalization_threshold = get_from_conf(conf, 'Frames',
            'normalization threshold', default_conf_obj.frames_normalization_threshold)
        self.frames_add_selection_dialog = get_from_conf(conf, 'Frames',
            'add selection dialog', default_conf_obj.frames_add_selection_dialog)

        self.align_frames_fast_changing_object = get_from_conf(conf, 'Align frames',
            'fast changing object', default_conf_obj.align_frames_fast_changing_object)
        self.align_frames_mode = get_from_conf(conf, 'Align frames',
            'mode', default_conf_obj.align_frames_mode)
        self.align_frames_automation = get_from_conf(conf, 'Align frames',
            'automation', default_conf_obj.align_frames_automation)
        self.align_frames_rectangle_scale_factor = get_from_conf(conf, 'Align frames',
            'rectangle scale factor', default_conf_obj.align_frames_rectangle_scale_factor)
        self.align_frames_search_width = get_from_conf(conf, 'Align frames',
            'search width', default_conf_obj.align_frames_search_width)
        self.align_frames_average_frame_percent = get_from_conf(conf, 'Align frames',
            'average frame percent', default_conf_obj.align_frames_average_frame_percent)

        self.alignment_points_half_box_width = get_from_conf(conf, 'Alignment points',
            'half box width', default_conf_obj.alignment_points_half_box_width)
        self.alignment_points_search_width = get_from_conf(conf, 'Alignment points',
            'search width', default_conf_obj.alignment_points_search_width)
        self.alignment_points_structure_threshold = get_from_conf(conf, 'Alignment points',
            'structure threshold', default_conf_obj.alignment_points_structure_threshold)
        self.alignment_points_brightness_threshold = get_from_conf(conf, 'Alignment points',
            'brightness threshold', default_conf_obj.alignment_points_brightness_threshold)
        self.alignment_points_frame_percent = get_from_conf(conf, 'Alignment points',
            'frame percent', default_conf_obj.alignment_points_frame_percent)
        self.alignment_points_frame_number = get_from_conf(conf, 'Alignment points',
            'frame number', default_conf_obj.alignment_points_frame_number)

        self.stack_frames_drizzle_factor_string = get_from_conf(conf, 'Stack frames',
            'drizzle factor string', default_conf_obj.stack_frames_drizzle_factor_string)

    def store_all_parameters_to_config_parser(self):
        """
        Write all variable parameters from the current configuration into a ConfigParser object.

        :return: ConfigParser object with all parameters
        """

        # Clear the ConfigParser object.
        sections = self.config_parser_object.sections()
        for section in sections:
            self.config_parser_object.remove_section(section)

        # Copy all current parameters from the current configuration into the ConfigParser object.
        self.config_parser_object.add_section('Hidden parameters')
        self.set_parameter('Hidden parameters', 'current directory',
                           self.hidden_parameters_current_dir)
        self.set_parameter('Hidden parameters', 'main window x0',
                           str(self.hidden_parameters_main_window_x0))
        self.set_parameter('Hidden parameters', 'main window y0',
                           str(self.hidden_parameters_main_window_y0))
        self.set_parameter('Hidden parameters', 'main window width',
                           str(self.hidden_parameters_main_window_width))
        self.set_parameter('Hidden parameters', 'main window height',
                           str(self.hidden_parameters_main_window_height))
        self.set_parameter('Hidden parameters', 'main window maximized',
                           str(self.hidden_parameters_main_window_maximized))
        self.config_parser_object.add_section('Global parameters')
        self.set_parameter('Global parameters', 'display quickstart',
                           str(self.global_parameters_display_quickstart))
        self.set_parameter('Global parameters', 'version', self.global_parameters_version)
        self.set_parameter('Global parameters', 'protocol level',
                           str(self.global_parameters_protocol_level))
        self.set_parameter('Global parameters', 'write protocol to file',
                           str(self.global_parameters_write_protocol_to_file))
        self.set_parameter('Global parameters', 'store protocol with result',
                           str(self.global_parameters_store_protocol_with_result))
        self.set_parameter('Global parameters', 'buffering level',
                           str(self.global_parameters_buffering_level))
        self.set_parameter('Global parameters', 'maximum memory active',
                           str(self.global_parameters_maximum_memory_active))
        self.set_parameter('Global parameters', 'maximum memory amount',
                           str(self.global_parameters_maximum_memory_amount))
        self.set_parameter('Global parameters', 'include postprocessing',
                           str(self.global_parameters_include_postprocessing))
        self.set_parameter('Global parameters', 'image format',
                           self.global_parameters_image_format)
        self.set_parameter('Global parameters', 'parameters in filename',
                           str(self.global_parameters_parameters_in_filename))
        self.set_parameter('Global parameters', 'stack number frames',
                           str(self.global_parameters_stack_number_frames))
        self.set_parameter('Global parameters', 'stack percent frames',
                           str(self.global_parameters_stack_percent_frames))
        self.set_parameter('Global parameters', 'ap box size',
                           str(self.global_parameters_ap_box_size))
        self.set_parameter('Global parameters', 'ap number',
                           str(self.global_parameters_ap_number))

        self.config_parser_object.add_section('Frames')
        self.set_parameter('Frames', 'gauss width', str(self.frames_gauss_width))
        self.set_parameter('Frames', 'debayering default', self.frames_debayering_default)
        self.set_parameter('Frames', 'debayering method', self.frames_debayering_method)
        self.set_parameter('Frames', 'normalization', str(self.frames_normalization))
        self.set_parameter('Frames', 'normalization threshold',
                           str(self.frames_normalization_threshold))
        self.set_parameter('Frames', 'add selection dialog', str(self.frames_add_selection_dialog))

        self.config_parser_object.add_section('Align frames')
        self.set_parameter('Align frames', 'fast changing object',
                           str(self.align_frames_fast_changing_object))
        self.set_parameter('Align frames', 'mode', self.align_frames_mode)
        self.set_parameter('Align frames', 'automation', str(self.align_frames_automation))
        self.set_parameter('Align frames', 'rectangle scale factor',
                           str(self.align_frames_rectangle_scale_factor))
        self.set_parameter('Align frames', 'search width',
                           str(self.align_frames_search_width))
        self.set_parameter('Align frames', 'average frame percent',
                           str(self.align_frames_average_frame_percent))

        self.config_parser_object.add_section('Alignment points')
        self.set_parameter('Alignment points', 'half box width',
                           str(self.alignment_points_half_box_width))
        self.set_parameter('Alignment points', 'search width',
                           str(self.alignment_points_search_width))
        self.set_parameter('Alignment points', 'structure threshold',
                           str(self.alignment_points_structure_threshold))
        self.set_parameter('Alignment points', 'brightness threshold',
                           str(self.alignment_points_brightness_threshold))
        self.set_parameter('Alignment points', 'frame percent',
                           str(self.alignment_points_frame_percent))
        self.set_parameter('Alignment points', 'frame number',
                           str(self.alignment_points_frame_number))

        self.config_parser_object.add_section('Stack frames')
        self.set_parameter('Stack frames', 'drizzle factor string',
                           self.stack_frames_drizzle_factor_string)

    def set_parameter(self, section, name, value):
        """
        Assign a new value to a parameter in the configuration object. The value is not checked for
        validity. Therefore, this method should be used with well-defined values internally only.

        :param section: section name (e.g. 'Global parameters') within the JSON data object
        :param name: name of the parameter (e.g. 'protocol level')
        :param value: new value to be assigned to the parameter (type str)
        :return: True, if the parameter was assigned successfully. False, otherwise.
        """

        try:
            self.config_parser_object.set(section, name, value)
            return True
        except:
            return False

    def set_derived_parameters(self):
        """
        Set parameters which are computed from other parameters.

        :return: -
        """

        # Set the alignment patch size to 1.5 times the box size.
        self.alignment_points_half_patch_width = int(
            round((self.alignment_points_half_box_width * 3) / 2))

        # Set the AP distance per coordinate direction such that adjacent patches overlap by 1/6
        # step on both sides.
        self.alignment_points_step_size = int(
            round((self.alignment_points_half_patch_width * 4.5) / 3))

        # Set the drizzling parameters.
        if self.stack_frames_drizzle_factor_string == "Off":
            self.drizzle_factor = 1
            self.drizzle_factor_is_1_5 = False
        elif self.stack_frames_drizzle_factor_string == "1.5x":
            self.drizzle_factor = 3
            self.drizzle_factor_is_1_5 = True
        elif self.stack_frames_drizzle_factor_string == "2x":
            self.drizzle_factor = 2
            self.drizzle_factor_is_1_5 = False
        elif self.stack_frames_drizzle_factor_string == "3x":
            self.drizzle_factor = 3
            self.drizzle_factor_is_1_5 = False

    def write_config(self, file_name=None):
        """
        Write all variable configuration parameters to a file. If no file name is specified, take
        the standard ".ini" file

        :param file_name: Optional configuration file name.
        :return: -
        """

        # Reset the ConfigParser object, and fill it with the current parameters.
        self.store_all_parameters_to_config_parser()

        # Add the contents of the postprocessing data object to the ConfigParser object.
        self.postproc_data_object.dump_config(self.config_parser_object)

        if not file_name:
            file_name = self.config_filename

        with open(file_name, 'w') as config_file:
            self.config_parser_object.write(config_file)

    def read_config(self, file_name=None):
        """
        Read the configuration from a file. If no name is given, the config is read from the
        standard ".ini" file in the user's home directory.

        :param file_name: Optional configuration file name
        :return: ConfigParser object with configuration parameters
        """

        # Initialize the ConfigParser object for parameters which the user can change.
        self.config_parser_object = ConfigParser()

        if not file_name:
            file_name = self.config_filename
        self.config_parser_object.read(file_name)

        # Get all stacking parameters from the parser object.
        self.get_all_parameters_from_configparser(self.config_parser_object)

        # Transfer all postprocessing parameters from the parser object to the postproc object.
        self.postproc_data_object.load_config(self.config_parser_object)

        return self.config_parser_object

def get_from_conf(config_parser_object, section, name, default):
    """
    Try to read a parameter from the config parser object. If it is not found, a default parameter
    is set instead.

    :param config_parser_object: Config parser object with all input parameters.
    :param section: Section name in config parser object.
    :param name: Variable name in section.
    :param default: Default value if variable cannot be read.
    :return: Either the value read, or default.
    """
    try:
        if isinstance(default, str):
            value = config_parser_object.get(section, name)
        elif isinstance(default, bool):
            value = config_parser_object.getboolean(section, name)
        elif isinstance(default, int):
            value = config_parser_object.getint(section, name)
        elif isinstance(default, float):
            value = config_parser_object.getfloat(section, name)
        else:
            raise ArgumentError("Parameter " + name + " in section " + section + " cannot be parsed")

    except:
        value = default

    return value

class PostprocDataObject(object):
    """
    This class implements the central data object used in postprocessing.

    """

    def __init__(self, postproc_suffix):
        """
        Initialize the data object.

        :param postproc_suffix: File suffix used for postprocessing result.
        :return: -
        """

        self.postproc_suffix = postproc_suffix

        # Initialize the postprocessing image versions with the unprocessed image (as version 0).
        self.initialize_versions()

        # Create a first processed version with initial parameters for Gaussian radius. The amount
        # of sharpening is initialized to zero.
        initial_version = self.add_postproc_version()
        initial_version.add_postproc_layer(PostprocLayer("Multilevel unsharp masking", 1., 1., 0.,
                                                         13, 0., False))

        # Initialize the pointer to the currently selected version to 0 (input image).
        # "version_compared" is used by the blink comparator later on. The blink comparator is
        # switched off initially.
        self.blinking = False
        self.version_compared = 0

    def set_postproc_input_image(self, image_original, name_original, image_format):
        """
        Set the input image and associated file name for postprocessing, and set derived variables.

        :param image_original: Image file (16bit Tiff) holding the input for postprocessing
        :param name_original: Path name of the original image.
        :param image_format: Image format, either 'tiff' or 'fits'.
        :return: -
        """

        self.image_original = image_original
        self.color = len(self.image_original.shape) == 3
        self.file_name_original = name_original

        # Set the standard path to the resulting image using the provided file suffix.
        self.file_name_processed = PostprocDataObject.set_file_name_processed(
            self.file_name_original, self.postproc_suffix, image_format)

        # Reset images and shift vectors which still might hold data from the previous job.
        for version in self.versions:
            version.set_image(self.image_original)
            version.input_image_saved = None
            version.input_image_hsv_saved = None
            version.shift_red_saved = None
            version.shift_blue_saved = None
            version.last_rgb_automatic = None

    @staticmethod
    def set_file_name_processed(file_name_original, postproc_suffix, image_format):
        """
        Derive the postprocessing output file name from the name of postprocessing input.

        :param file_name_original: Postprocessing input file name (e.g. result from stacking)
        :param postproc_suffix: Additional suffix to be inserted before file extension.
        :param image_format: Image format, either 'png', 'tiff' or 'fits'.
        :return: Name of postprocessing result.
        """

        return splitext(file_name_original)[0] + postproc_suffix + '.' + image_format

    def initialize_versions(self):
        """
        Initialize the postprocessing image versions with the unprocessed image (as version 0).

        :return: -
        """

        original_version = PostprocVersion()
        self.versions = [original_version]
        self.number_versions = 0
        self.version_selected = 0

    def add_postproc_version(self):
        """
        Add a new postprocessing version, and set the "selected" pointer to it.

        :return: The new version object
        """

        new_version = PostprocVersion()
        self.versions.append(new_version)
        self.number_versions += 1
        self.version_selected = self.number_versions
        return new_version

    def new_postproc_version_from_existing(self):
        """
        Create a new postprocessing version by copying the currently selected one. Append the new
        version to the list of postprocessing versions. Set the current version pointer to the new
        version.
        :return: New postprocessing version.
        """

        new_version = deepcopy(self.versions[self.version_selected])
        self.versions.append(new_version)
        self.number_versions += 1
        self.version_selected = self.number_versions
        return new_version

    def remove_postproc_version(self, index):
        """
        Remove a postprocessing version, and set the "selected" pointer to the previous one. The
        initial version (input image) cannot be removed.

        :param index: Index of the version to be removed.
        :return: -
        """

        if 0 < index <= self.number_versions:
            self.versions = self.versions[:index] + self.versions[index + 1:]
            self.number_versions -= 1
            self.version_selected = index - 1

    def finalize_postproc_version(self, version_index=None):
        """
        If "correction mode" is on, the image stored with the selected version is in 8bit and
        potentially interpolated by the resolution factor. Before the image is saved, it has to be
        processed from the original image using the current shift and layer data. The 16bit image is
        stored with the current version object.

        :param version_index: If an index is specified, process the version with this index.
                              Otherwise, process the version currently selected.
        :return: The processed 16bit image (also stored with the version).
        """

        if version_index is None:
            # The computation is based on the current "version selected".
            version = self.versions[self.version_selected]
        else:
            version = self.versions[version_index]

        # If the dtype is "uint16", nothing is to be done. If it is "uint8", the current image is
        # an intermediate product of the correction mode. In this case the rigorous postprocessing
        # pipeline must be traversed to get the final image.
        if version.image is not None and version.image.dtype == uint8:

            # Compute the complete current shifts, including temporary shift corrections.
            shift_red = (version.shift_red[0] + version.correction_red[0],
                        version.shift_red[1] + version.correction_red[1])
            shift_blue = (version.shift_blue[0] + version.correction_blue[0],
                        version.shift_blue[1] + version.correction_blue[1])

            # Shift the image with the resolution given by the selected interpolation factor.
            interpolation_factor = [1, 2, 4][version.rgb_resolution_index]
            shifted_image = Miscellaneous.shift_colors(self.image_original,
                                                       shift_red, shift_blue,
                                                       interpolate_input=interpolation_factor,
                                                       reduce_output=interpolation_factor)

            # Apply all wavelet layers.
            version.image = Miscellaneous.post_process(shifted_image, version.layers)

        return version.image

    def dump_config(self, config_parser_object):
        """
        Dump all version and layer parameters into a ConfigParser object (except for image data).
        Version 0 is not included because it does not contain any layer info.

        :param config_parser_object: ConfigParser object.
        :return: -
        """

        # First remove old postprocessing sections, if there are any in the ConfigParser object.
        for section in config_parser_object.sections():
            section_items = section.split()
            if section_items[0] == 'PostprocessingVersion' or \
                    section_items[0] == 'PostprocessingInfo':
                config_parser_object.remove_section(section)

        # Store general postprocessing info.
        config_parser_object.add_section('PostprocessingInfo')
        config_parser_object.set('PostprocessingInfo', 'version selected',
                                 str(self.version_selected))

        # For every version create a section with RGB alignment parameters.
        for version_index, version in enumerate(self.versions):
            section_name = "PostprocessingVersion " + str(version_index) + " RGBAlignment"
            config_parser_object.add_section(section_name)
            # Add the parameters of this version.
            config_parser_object.set(section_name, 'rgb automatic', str(version.rgb_automatic))
            config_parser_object.set(section_name, 'rgb gauss width', str(version.rgb_gauss_width))
            config_parser_object.set(section_name, 'rgb resolution index',
                                     str(version.rgb_resolution_index))
            config_parser_object.set(section_name, 'rgb shift red y',
                                     str(version.shift_red[0]))
            config_parser_object.set(section_name, 'rgb shift red x',
                                     str(version.shift_red[1]))
            config_parser_object.set(section_name, 'rgb shift blue y',
                                     str(version.shift_blue[0]))
            config_parser_object.set(section_name, 'rgb shift blue x',
                                     str(version.shift_blue[1]))

            # For every postprocessing layer of this version, create a separate section.
            for layer_index, layer in enumerate(version.layers):
                section_name = "PostprocessingVersion " + str(version_index) + " Layer " + str(
                    layer_index)
                config_parser_object.add_section(section_name)

                # Add the parameters of the layer.
                config_parser_object.set(section_name, 'postprocessing method', layer.postproc_method)
                config_parser_object.set(section_name, 'radius', str(layer.radius))
                config_parser_object.set(section_name, 'amount', str(layer.amount))
                config_parser_object.set(section_name, 'bilateral fraction', str(layer.bi_fraction))
                config_parser_object.set(section_name, 'bilateral range', str(layer.bi_range))
                config_parser_object.set(section_name, 'denoise', str(layer.denoise))
                config_parser_object.set(section_name, 'luminance only', str(layer.luminance_only))



    def load_config(self, config_parser_object):
        """
        Load all postprocessing configuration data from a ConfigParser object. The data replace
        all versions (apart from version 0) and all associated layer info. The image data is taken
        from the current data object and is not restored.

        :param config_parser_object: ConfigParser object.
        :return: -
        """

        standard_version = PostprocVersion()
        standard_layer = PostprocLayer("Multilevel unsharp masking", 1., 1., 0., 20, 0., False)

        # Initialize the postprocessing image versions with the unprocessed image (as version 0).
        self.initialize_versions()

        # Load general postprocessing info.
        try:
            self.version_selected = config_parser_object.getint('PostprocessingInfo',
                                                                'version selected')
        except:
            # the ConfigParser object does not contain postprocessing info. Leave the data object
            # with only the original version stored.
            return

        # Initialize the version index for comparison with an impossible value.
        self.number_versions = 0

        # Go through all sections and find the postprocessing sections.
        for section in config_parser_object.sections():
            section_items = section.split()
            if section_items[0] == 'PostprocessingVersion':

                # For a new version the first section contains the RGB alignment info. Create a new
                # version and store the RGB alignment parameters with it. For the first (neutral)
                # version 0 only RGB shift parameters are stored (no layers). In this case
                # initialize the version list.
                if section_items[2] == 'RGBAlignment':
                    version_index = int(section_items[1])
                    if version_index:
                        new_version = self.add_postproc_version()
                    else:
                        self.initialize_versions()
                        new_version = self.versions[0]
                    new_version.rgb_automatic = get_from_conf(config_parser_object, section,
                        'rgb automatic', standard_version.rgb_automatic)
                    new_version.rgb_gauss_width = get_from_conf(config_parser_object, section,
                        'rgb gauss width', standard_version.rgb_gauss_width)
                    new_version.rgb_resolution_index = get_from_conf(config_parser_object, section,
                        'rgb resolution index', standard_version.rgb_resolution_index)
                    new_version.shift_red = (get_from_conf(config_parser_object, section,
                        'rgb shift red y', standard_version.shift_red[0]),
                        get_from_conf(config_parser_object, section, 'rgb shift red x',
                        standard_version.shift_red[1]))
                    new_version.shift_blue = (get_from_conf(config_parser_object, section,
                        'rgb shift blue y', standard_version.shift_blue[0]),
                        get_from_conf(config_parser_object, section, 'rgb shift blue x',
                        standard_version.shift_blue[1]))

                # A layer section is found. Store it for the current version.
                elif section_items[2] == 'Layer':
                    # Read all parameters of this layer, and add a layer to the current version.
                    method = get_from_conf(config_parser_object, section,
                        'postprocessing method', standard_layer.postproc_method)
                    radius = get_from_conf(config_parser_object, section,
                        'radius', standard_layer.radius)
                    amount = get_from_conf(config_parser_object, section,
                        'amount', standard_layer.amount)
                    bi_fraction = get_from_conf(config_parser_object, section,
                        'bilateral fraction', standard_layer.bi_fraction)
                    bi_range = get_from_conf(config_parser_object, section,
                        'bilateral range', standard_layer.bi_range)
                    denoise = get_from_conf(config_parser_object, section,
                        'denoise', standard_layer.denoise)
                    luminance_only = get_from_conf(config_parser_object, section,
                        'luminance only', standard_layer.luminance_only)
                    new_version.add_postproc_layer(PostprocLayer(method, radius, amount, bi_fraction,
                                                                 bi_range, denoise, luminance_only))

        # Set the selected version again, because not all versions might have been read successfully.
        self.version_selected = min(config_parser_object.getint('PostprocessingInfo',
                                                            'version selected'), self.number_versions)


class PostprocVersion(object):
    """
    Instances of this class hold the data defining a single postprocessing version, including the
    resulting image for the current parameter set.
    """

    def __init__(self):
        """
        Initialize the version object with the input image and an empty set of processing layers.
        :param image: Input image (16bit Tiff) for postprocessing
        """

        self.postproc_method = "Multilevel unsharp masking"
        self.layers = []
        self.number_layers = 0
        self.rgb_automatic = False
        self.last_rgb_automatic = None
        self.rgb_gauss_width = 7
        self.rgb_resolution_index = 1
        self.shift_red = (0., 0.)
        self.shift_blue = (0., 0.)
        self.shift_red_saved = None
        self.shift_blue_saved = None
        self.rgb_correction_mode = False
        self.correction_red = (0., 0.)
        self.correction_blue = (0., 0.)
        self.correction_red_saved = (0., 0.)
        self.correction_blue_saved = (0., 0.)
        self.image = None
        self.images_uncorrected = [None] * 3
        self.input_image_saved = None
        self.input_image_hsv_saved = None

    def set_image(self, image):
        """
        Set the current image for this version.

        :param image: Image file (16bit Tiff) holding the pixel data of this version's image.
        :return: -
        """

        self.image = image

    def add_postproc_layer(self, layer):
        """
        Add a postprocessing layer.
        :param layer: Layer instance to be added to the list of layers.
        :return: -
        """

        self.layers.append(layer)
        self.number_layers += 1

    def remove_postproc_layer(self, layer_index):
        """
        Remove a postprocessing layer from this version. If there is only one layer, do not delete
        it, but reset it to standard values instead. This makes sure a version always has at least
        one layer.

        :param layer_index: Index of the layer to be removed.
        :return: -
        """

        if self.number_layers == 1:
            self.layers = [PostprocLayer("Multilevel unsharp masking", 1., 1., 0., 20, 0., False)]
        else:
            if 0 <= layer_index < self.number_layers:
                self.layers = self.layers[:layer_index] + self.layers[layer_index + 1:]
                self.number_layers -= 1


class PostprocLayer(object):
    """
    Instances of this class hold the parameters which define a postprocessing layer.
    """

    def __init__(self, method, radius, amount, bi_fraction, bi_range, denoise, luminance_only):
        """
        Initialize the Layer instance with values for Gaussian radius, amount of sharpening and a
        flag which indicates on which channel the sharpening is to be applied.

        :param method: Description of the sharpening method.
        :param radius: Radius (in pixels) of the Gaussian sharpening kernel.
        :param amount: Amount of sharpening for this layer.
        :param bi_fraction: Fraction of bilateral vs. Gaussian filter
                            (0.: only Gaussian, 1.: only bilateral).
        :param bi_range: luminosity range parameter of bilateral filter (0 <= bi_range <= 255).
                         Please note that for 16bit images the true range is 256 times as high.
        :param denoise: Fraction of Gaussian blur to be applied to this layer
                        (0.: No Gaussian blur, 1.: Full filter application).
        :param luminance_only: True, if sharpening is to be applied to the luminance channel only.
                               False, otherwise.
        """

        self.postproc_method = method
        self.radius = radius
        self.amount = amount
        self.bi_fraction = bi_fraction
        self.bi_range = bi_range
        self.denoise = denoise
        self.luminance_only = luminance_only



# -*- coding: utf-8 -*-

from sys import argv

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt

from list_view_widget import Ui_ListViewWidget


class ListView(QtWidgets.QWidget, Ui_ListViewWidget):
    def __init__(self):
        """
        Initialization of the widget.
        """
        super(ListView, self).__init__()
        self.setupUi(self)

        self.frames_included = []
        self.items_selected = None

        self.background_included = QtGui.QColor(130, 255, 130)
        self.background_excluded = QtGui.QColor(120, 120, 120)

        self.listWidget.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

        for i in range(30):
            item = QtWidgets.QListWidgetItem("Frame %i included" % i)
            item.setBackground(self.background_included)
            item.setForeground(QtGui.QColor(0, 0, 0))
            self.listWidget.addItem(item)
            self.frames_included.append(True)

        self.indices_selected = []

        self.listWidget.installEventFilter(self)
        self.listWidget.itemClicked.connect(self.select_items)
        self.addButton.clicked.connect(self.use_triggered)
        self.removeButton.clicked.connect(self.not_use_triggered)

    def select_items(self):
        self.items_selected = self.listWidget.selectedItems()
        self.indices_selected = [self.listWidget.row(item) for item in self.items_selected]
        print(self.indices_selected)

    def eventFilter(self, source, event):
        if source is self.listWidget:
            if event.type() == QtCore.QEvent.ContextMenu:
                print("Context menu opened")
                menu = QtWidgets.QMenu()
                action1 = QtWidgets.QAction('Use for stacking', menu)
                action1.triggered.connect(self.use_triggered)
                menu.addAction((action1))
                action2 = QtWidgets.QAction("Don't use for stacking", menu)
                action2.triggered.connect(self.not_use_triggered)
                menu.addAction((action2))
                menu.exec_(event.globalPos())
            elif event.type() == QtCore.QEvent.KeyPress:
                if event.key() == Qt.Key_Plus:
                    self.use_triggered()
                elif event.key() == Qt.Key_Minus:
                    self.not_use_triggered()
                elif event.key() == Qt.Key_Escape:
                    self.items_selected = []
                    self.indices_selected = []
        return super(ListView, self).eventFilter(source, event)

    def use_triggered(self):
        if self.items_selected:
            for index, item in enumerate(self.items_selected):
                index_selcted = self.indices_selected[index]
                item.setText("Frame %i included" % index_selcted)
                item.setBackground(self.background_included)
                item.setForeground(QtGui.QColor(0, 0, 0))
                self.frames_included[index_selcted] = True

            print(str(self.frames_included))

    def not_use_triggered(self):
        if self.items_selected:
            for index, item in enumerate(self.items_selected):
                index_selcted = self.indices_selected[index]
                item.setText("Frame %i excluded" % index_selcted)
                item.setBackground(self.background_excluded)
                item.setForeground(QtGui.QColor(255, 255, 255))
                self.frames_included[index_selcted] = False

            print(str(self.frames_included))


if __name__ == '__main__':
    app = QtWidgets.QApplication(argv)
    window = ListView()
    window.show()
    app.exec_()



# -*- coding: utf-8; -*-
"""
Copyright (c) 2019 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

"""

from copy import deepcopy
from pathlib import Path

from PyQt5 import QtWidgets, QtCore, QtGui

from exceptions import InternalError
from job_dialog import Ui_JobDialog

# The following lists define the allowed file extensions for still images and videos.
image_extensions = ['.tif', '.tiff', '.fit', '.fits', '.jpg', '.png']
video_extensions = ['.avi', '.mov', '.mp4', '.ser']
extensions = image_extensions + video_extensions


class FileDialog(QtWidgets.QFileDialog):
    """
    This is a variant of the regular FileDialog class which allows to choose files and directories
    at the same time.
    """

    signal_dialog_ready = QtCore.pyqtSignal(list)

    def __init__(self, *args, **kwargs):
        super(FileDialog, self).__init__(*args, **kwargs)
        # Do not use the native dialog of the OS. Otherwise the selection model tree is not
        # available as expected.
        self.setOption(QtWidgets.QFileDialog.DontUseNativeDialog, True)
        self.setFileMode(QtWidgets.QFileDialog.ExistingFiles)
        self.tree = self.findChild(QtWidgets.QTreeView)

    def accept(self):
        """
        When the selection of files and directories is acknowledged, construct a list of strings
        with their names and send them via a signal to the job editor.

        :return: -
        """

        files = [self.directory().filePath(str(i.data())) for i in
                 self.tree.selectionModel().selectedIndexes() if i.column() == 0]
        self.signal_dialog_ready.emit(files)
        self.close()


class Job(object):
    """
    Objects of this class encapsulate all information describing a PSS job.
    """

    def __init__(self, job_name):
        """
        Initialize a Job object, given its name. The following instance variables are set:
           - name: Path name string coming from the file chooser.
           - file_name: File name string without path.
           - type: Either 'video', or 'image' for stacking jobs, or 'postproc' for postprocessing.
           - bayer_option_selected: Initialized to 'Auto detect color' for file types for which
                                    debayering is supported. Otherwise None.
           - bayer_pattern: Initialized to None

        :param job_name: Name of the job (str)
        """

        self.name = job_name
        path = Path(self.name)
        self.file_name = path.name

        # Bayer patterns are only defined for type 'video'.
        self.bayer_pattern = None
        self.bayer_option_selected = None

        # Set the type of the job based on the file name extension.
        if path.is_file():
            extension = path.suffix.lower()
            if extension in video_extensions:
                self.type = 'video'
                self.bayer_option_selected = 'Auto detect color'
            elif extension in image_extensions:
                self.type = 'postproc'
            else:
                raise InternalError(
                    "Unsupported file type '" + extension + "' specified for job")
        elif path.is_dir():
            self.type = 'image'
            self.bayer_option_selected = 'Auto detect color'
        else:
            raise InternalError("Cannot decide if input file is video or image directory")


class JoblistWidget(QtWidgets.QListWidget):
    """
    This is a customized version of the standard QListWidget which accepts "drag and drop" of
    picture files and directories. Dropped items are sent to the higher-level class as a list
    via the signal "signal_job_entries".
    """

    signal_job_entries = QtCore.pyqtSignal(list)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setAcceptDrops(True)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        if event.mimeData().hasUrls():
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """
        Handle a drop event if its mime type fits.

        :param event: Drop event
        :return: -
        """

        if event.mimeData().hasUrls():
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()

            # Construct a list with the job names received by the drop event.
            links = []
            for url in event.mimeData().urls():
                # Only local files (with appropriate extensions) and directories are accepted.
                if url.isLocalFile():
                    file_path = Path(str(url.toLocalFile()))
                    if file_path.is_dir() or file_path.suffix in extensions:
                        links.append(str(url.toLocalFile()))
            # Send the job name list to the upper-level method "get_input_names".
            self.signal_job_entries.emit(links)
        else:
            event.ignore()


class JobEditor(QtWidgets.QFrame, Ui_JobDialog):
    """
    Manage the list of jobs. Each item is either the name of a video file (.avi .mov .mp4 .ser) or a directory
    containing image files of the same shape. Ask the user to add jobs to the list, or to remove
    existing entries. The interaction with the user is through the JobDialog class.
    """

    def __init__(self, parent_gui, parent=None):
        """
        Initialize the job editor. The widget has a fixed size and is rendered as a QFrame.

        :param parent_gui: GUI object by which the editor is invoked.
        :param parent: Parent object
        """

        QtWidgets.QFrame.__init__(self, parent)
        self.setupUi(self)
        self.setWindowIcon(QtGui.QIcon('../PSS-Icon-64.ico'))

        # Replace the standard QListWidget with a customized version which implements
        # "drag and drop". It sends dropped items via its signal "signal_job_entries".
        self.job_list_widget = JoblistWidget()
        self.gridLayout.addWidget(self.job_list_widget, 0, 0, 1, 6)
        self.job_list_widget.signal_job_entries.connect(self.get_input_names)

        self.setFrameShape(QtWidgets.QFrame.Panel)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.setObjectName("configuration_editor")

        # The following line was deactivated. Otherwise the instructions under the joblist
        # would not show completely on full HD monitors.
        # self.setFixedSize(900, 600)

        self.parent_gui = parent_gui
        self.configuration = self.parent_gui.configuration

        # De-activate main gui menu entries while the job editor is active. They will be
        # re-activated upon leaving the editor with "OK" or "Cancel".
        self.parent_gui.activate_gui_elements(
            [self.parent_gui.ui.menuFile, self.parent_gui.ui.menuEdit,
             self.parent_gui.ui.menuCalibrate], False)

        # Set the window icon to the PSS icon.
        self.setWindowIcon(QtGui.QIcon(self.configuration.window_icon))

        # Get a copy of the jobs constructed so far. The editor only works on the copy, so
        # that in the case of "cancel" the original list is not changed.
        self.jobs = deepcopy(self.parent_gui.jobs)

        self.messageLabel.setStyleSheet('color: red')

        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.button_remove_jobs.clicked.connect(self.remove_job_list)
        self.button_add_jobs.clicked.connect(self.add_jobs)

        # Install an event filter on the job list to enable the context menu.
        self.job_list_widget.installEventFilter(self)

        # Populate the job list widget with the current job list.
        self.populate_job_list()

        # If the job list is empty, open the input file dialog.
        if not self.jobs:
            self.add_jobs()

    def populate_job_list(self):
        """
        Fill the central QListWidget with the current list of job names.

        :return: -
        """
        self.job_list_widget.clear()
        for job in self.jobs:
            item = QtWidgets.QListWidgetItem(job.name)
            self.job_list_widget.addItem(item)

    def add_jobs(self):
        """
        Open a file dialog for entering additional job names. Entries can either be video files or
        directories for stacking, or single image files for postprocessing.

        :return: -
        """

        options = QtWidgets.QFileDialog.Options()
        message = "Select video file(s)/folders with image files for stacking, and/or " \
                  "image files for postprocessing"

        self.file_dialog = FileDialog(self, message,
                                      self.configuration.hidden_parameters_current_dir,
                                      "Videos (*.avi *.mov *.mp4 *.ser)", options=options)
        self.file_dialog.setNameFilters(["Still image folders / video files for stacking (*.avi *.mov *.mp4 *.ser)",
                                         "Images for postprocessing (*.tiff *.tif *.fit *.fits *.png *.jpg)"])
        self.file_dialog.selectNameFilter("Still image folders / video files for stacking (*.avi *.mov *.mp4 *.ser)")

        # The list of strings with the new job names is sent by the FileDialog via the signal.
        self.file_dialog.signal_dialog_ready.connect(self.get_input_names)
        self.file_dialog.exec_()

    def get_input_names(self, input_names):
        """
        Receive the list of new job names from the FileDialog.

        :param input_names: List of strings with job names
        :return: -
        """
        if input_names:
            for input_name in input_names:
                if input_name not in [job.name for job in self.jobs]:
                    try:
                        new_job = Job(input_name)
                        new_job.bayer_option_selected = self.configuration.frames_debayering_default
                        self.jobs.append(new_job)
                    except:
                        pass

            # Save the current directory location. The next dialog will open at this position.
            self.configuration.hidden_parameters_current_dir = str(Path(input_names[0]).parents[0])
        self.populate_job_list()

    def remove_job_list(self):
        """
        Remove the job entries which are currently selected.

        :return: -
        """

        # Get the selected items from the central job list widget.
        remove_list = [str(item.text()) for item in self.job_list_widget.selectedItems()]

        # Update the current job list, and re-draw the job list widget.
        self.jobs = [job for job in self.jobs if job.name not in remove_list]
        self.populate_job_list()

    def eventFilter(self, source, event):
        """
        The event filter intercepts context menu events on job list items. It is used to specify
        Bayer patterns explicitly.

        :param source: The widget for which the filter is activated
                       (in this case "self.job_list_widget")
        :param event: The event type for this filter ("QtCore.QEvent.ContextMenu")
        :return: True
        """

        # Restrict the selected items to the subset for which debayering is supported.
        selected_items = [item for item in self.job_list_widget.selectedItems() if
                          self.jobs[source.row(item)].type == 'video']

        self.pattern = None

        # If a context menu item is pressed, remember the pattern.
        def action1_triggered(state):
            self.pattern = 'Auto detect color'

        def action2_triggered(state):
            self.pattern = 'RGB'

        def action3_triggered(state):
            self.pattern = 'BGR'

        def action4_triggered(state):
            self.pattern = 'Grayscale'

        def action5_triggered(state):
            self.pattern = 'Force Bayer RGGB'

        def action6_triggered(state):
            self.pattern = 'Force Bayer GRBG'

        def action7_triggered(state):
            self.pattern = 'Force Bayer GBRG'

        def action8_triggered(state):
            self.pattern = 'Force Bayer BGGR'

        # The context menu is opened on a job list entry.
        if (event.type() == QtCore.QEvent.ContextMenu and
                source is self.job_list_widget):

            # Show the context menu only if job items are selected for which debayering is
            # supported.
            if selected_items:
                # Create a list of patterns which are checked by the selected items initially.
                checked_patterns = [self.jobs[source.row(item)].bayer_option_selected
                                    for item in selected_items]

                # Create the context menu. Mark those patterns checked which have been set for at least
                # one selected job list entry.
                menu = QtWidgets.QMenu()
                action1 = QtWidgets.QAction('Auto detect color', menu, checkable=True)
                action1.triggered.connect(action1_triggered)
                if 'Auto detect color' in checked_patterns:
                    action1.setChecked(True)
                menu.addAction(action1)
                menu.addSeparator()
                action2 = QtWidgets.QAction('RGB', menu, checkable=True)
                action2.triggered.connect(action2_triggered)
                if 'RGB' in checked_patterns:
                    action2.setChecked(True)
                menu.addAction(action2)
                action3 = QtWidgets.QAction('BGR', menu, checkable=True)
                action3.triggered.connect(action3_triggered)
                if 'BGR' in checked_patterns:
                    action3.setChecked(True)
                menu.addAction(action3)
                action4 = QtWidgets.QAction('Grayscale', menu, checkable=True)
                action4.triggered.connect(action4_triggered)
                if 'Grayscale' in checked_patterns:
                    action4.setChecked(True)
                menu.addAction(action4)
                action5 = QtWidgets.QAction('Force Bayer RGGB', menu, checkable=True)
                action5.triggered.connect(action5_triggered)
                if 'Force Bayer RGGB' in checked_patterns:
                    action5.setChecked(True)
                menu.addAction(action5)
                action6 = QtWidgets.QAction('Force Bayer GRBG', menu, checkable=True)
                action6.triggered.connect(action6_triggered)
                if 'Force Bayer GRBG' in checked_patterns:
                    action6.setChecked(True)
                menu.addAction(action6)
                action7 = QtWidgets.QAction('Force Bayer GBRG', menu, checkable=True)
                action7.triggered.connect(action7_triggered)
                if 'Force Bayer GBRG' in checked_patterns:
                    action7.setChecked(True)
                menu.addAction(action7)
                action8 = QtWidgets.QAction('Force Bayer BGGR', menu, checkable=True)
                action8.triggered.connect(action8_triggered)
                if 'Force Bayer BGGR' in checked_patterns:
                    action8.setChecked(True)
                menu.addAction(action8)

                # Identify the selected items and their locations in the job list. Set the selected
                # Bayer pattern in the corresponding job objects.
                if menu.exec_(event.globalPos()) and self.pattern is not None:
                    for item in selected_items:
                        row = source.row(item)
                        self.jobs[row].bayer_option_selected = self.pattern
                return True

        return super(JobEditor, self).eventFilter(source, event)

    def keyPressEvent(self, event):
        """
        The + and - keys are used for zooming.

        :param event: event object
        :return: -
        """

        # If the "+" key is pressed, zoom in. If "-" is pressed, zoom out.
        if event.key() == QtCore.Qt.Key_Delete:
            self.remove_job_list()
        else:
            super(JobEditor, self).keyPressEvent(event)

    def accept(self):
        """
        If the OK button is clicked and the job list has been changed update the job list in the
        parent object.

        :return: -
        """

        # Update the job list and reset the current job index to the first entry.
        self.parent_gui.jobs = self.jobs
        self.parent_gui.job_number = len(self.jobs)
        self.parent_gui.job_index = 0
        self.parent_gui.activity = "Read frames"

        # Re-activate the main GUI menus, and activate the "Automatic" checkbox.
        self.parent_gui.activate_gui_elements(
            [self.parent_gui.ui.menuFile, self.parent_gui.ui.menuEdit,
             self.parent_gui.ui.menuCalibrate, self.parent_gui.ui.box_automatic], True)
        self.parent_gui.update_status()
        self.close()

    def reject(self):
        """
        The Cancel button is pressed, discard the changes and close the GUI window.
        :return: -
        """

        # Re-activate the main GUI menus.
        self.parent_gui.activate_gui_elements(
            [self.parent_gui.ui.menuFile, self.parent_gui.ui.menuEdit,
             self.parent_gui.ui.menuCalibrate], True)
        self.close()

    def closeEvent(self, event):
        """
        Remove the job editor widget from the parent GUI and close the editor.

        :param event: Close event object
        :return: -
        """

        self.parent_gui.display_widget(None, display=False)
        self.close()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gui_with_label.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(599, 451)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setPointSize(10)
        self.label.setFont(font)
        self.label.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.label.setWordWrap(True)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.dont_show_checkBox = QtWidgets.QCheckBox(Form)
        self.dont_show_checkBox.setObjectName("dont_show_checkBox")
        self.verticalLayout.addWidget(self.dont_show_checkBox)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "TextLabel"))
        self.dont_show_checkBox.setText(_translate("Form", "Don\'t show this message again"))



# -*- coding: utf-8; -*-
"""
Copyright (c) 2019 Rolf Hempel, rolf6419@gmx.de

This file is part of the PlanetarySystemStacker tool (PSS).
https://github.com/Rolf-Hempel/PlanetarySystemStacker

PSS is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PSS.  If not, see <http://www.gnu.org/licenses/>.

Part of this module (in class "AlignmentPointEditor" was copied from
https://stackoverflow.com/questions/35508711/how-to-enable-pan-and-zoom-in-a-qgraphicsview

"""
# The following PyQt5 imports must precede any matplotlib imports. This is a workaround
# for a Matplotlib 2.2.2 bug.

from glob import glob
from sys import argv, exit
from time import time, sleep

import matplotlib

matplotlib.use('Agg')
import matplotlib.pyplot as plt
from PyQt5 import QtCore, QtGui, QtWidgets
from cv2 import NORM_MINMAX, normalize, cvtColor, COLOR_GRAY2RGB, circle, line
from matplotlib import patches
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as Canvas
from matplotlib.figure import Figure
from numpy import array, full, uint8, uint16, float64

from align_frames import AlignFrames
from configuration import Configuration
from exceptions import NotSupportedError, InternalError, Error
from frame_viewer_gui import Ui_frame_viewer
from frames import Frames
from miscellaneous import Miscellaneous
from rank_frames import RankFrames


class MatplotlibWidget(Canvas):
    """
    This widget creates a plot of frame qualities, either sorted chronologically or by quality.

    """

    def __init__(self, configuration, rank_frames, parent=None):
        """
        Initialize the widget.

        :param configuration: Configuration object with parameters
        :param rank_frames: RankFrames object with global quality ranks (between 0. and 1.,
                            1. being optimal) for all frames
        :param parent: Parent object
        """

        super(MatplotlibWidget, self).__init__(Figure())

        self.setParent(parent)
        self.configuration = configuration
        self.rank_frames = rank_frames
        self.line_chronological = None
        self.line_quality = None
        self.dot = None
        self.line_quality_cutoff = None
        self.patch_quality_cutoff = None

        plt.rcParams.update({'font.size': 8})

        self.fig, self.ax = plt.subplots()
        self.ax.invert_xaxis()
        plt.subplots_adjust(left=0.23, right=0.95, top=0.98, bottom=0.12)

    def renew_plot(self, index, frame_ordering, alignment_points_frame_number):
        """
        Update the complete plot, including the quality line, the cutoff line, and the dot at the
        position of the current frame.

        :param index: Index of the current frame (relative to the selected order)
        :param frame_ordering: Ordering of frames, either "chronological" or "quality"
        :param alignment_points_frame_number: Number of frames to be stacked at each AP.
        :return: -
        """

        self.ax.clear()
        self.line_quality_cutoff = None
        self.patch_quality_cutoff = None

        # The quality axis starts with 1. (highest quality).
        self.ax.invert_xaxis()

        # Remember objects drawn into the plot.
        self.line_chronological = None
        self.line_quality = None
        self.dot = None
        self.line_quality_cutoff = None

        # Plot the new data.
        self.plot_data(frame_ordering)
        self.plot_cutoff_lines(frame_ordering, alignment_points_frame_number)
        self.plot_dot(index)

    def plot_data(self, frame_ordering):
        """
        Plot the quality line. Frames (y axis) are ordered as specified with parameter
         "frame_ordering".

        :param frame_ordering: Ordering of frames, either "chronological" or "quality"
        :return: -
        """

        self.ax.set_ylim(0, self.rank_frames.number + 1)
        self.y = array(range(1, self.rank_frames.number + 1))
        if frame_ordering == "chronological":
            if self.line_chronological is not None:
                self.line_chronological.remove()
            self.x = array(self.rank_frames.frame_ranks)
            plt.ylabel('Frame numbers ordered chronologically')
            plt.gca().invert_yaxis()
            plt.xlabel('Quality')
            self.line_chronological, = plt.plot(self.x, self.y, lw=1, color='blue')
            plt.grid(True)
        else:
            if self.line_quality is not None:
                self.line_quality.remove()
            self.x = array(
                [self.rank_frames.frame_ranks[i] for i in self.rank_frames.quality_sorted_indices])
            plt.ylabel('Frame numbers ordered by quality')
            plt.gca().invert_yaxis()
            plt.xlabel('Quality')
            self.line_quality, = plt.plot(self.x, self.y, lw=1, color='green')
            plt.grid(True)
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def plot_cutoff_lines(self, frame_ordering, alignment_points_frame_number):
        """
        Plot a horizontal (if frames are ordered by quality) or vertical (if frames are ordered
        chronologically) line to separate frames to be stacked from the other ones.

        :param frame_ordering: Ordering of frames, either "chronological" or "quality"
        :param alignment_points_frame_number: Number of frames to be stacked at each AP.
        :return: -
        """
        if frame_ordering == "chronological":
            if self.line_quality_cutoff is not None:
                self.line_quality_cutoff.remove()
            if self.patch_quality_cutoff is not None:
                self.patch_quality_cutoff.remove()
            quality_cutoff = self.rank_frames.frame_ranks[
                self.rank_frames.quality_sorted_indices[alignment_points_frame_number-1]]
            x_cutoff = full((self.rank_frames.number,), quality_cutoff)
            self.line_quality_cutoff, = plt.plot(x_cutoff, self.y, lw=1, color='orange')
            width = 1. - quality_cutoff
            height = self.rank_frames.number - 1
            xy = (quality_cutoff, 1.)
        else:
            if self.line_quality_cutoff is not None:
                self.line_quality_cutoff.remove()
            if self.patch_quality_cutoff is not None:
                self.patch_quality_cutoff.remove()
            y_cutoff = full((self.rank_frames.number,), alignment_points_frame_number)
            self.line_quality_cutoff, = plt.plot(self.x, y_cutoff, lw=1, color='orange')
            width = 1. - self.x[-1]
            height = alignment_points_frame_number
            xy = (self.x[-1], 0.)
        self.patch_quality_cutoff = patches.Rectangle(xy, width, height, linewidth=None,
                                                      facecolor=(0.5, 0.2, 0., 0.15))
        self.ax.add_patch(self.patch_quality_cutoff)
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def plot_dot(self, frame_index):
        """
        Plot a dot on the quality line at the position of the current frame.

        :param frame_index: Frame index
        :return: -
        """

        if self.dot is not None:
            self.dot.remove()
        self.dot = plt.scatter(self.x[frame_index], self.y[frame_index], s=20, color='red')
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()


class FrameViewer(QtWidgets.QGraphicsView):
    """
    This widget implements a frame viewer. Panning and zooming is implemented by using the mouse
    and scroll wheel.

    """

    zoom_factor_signal = QtCore.pyqtSignal(int)

    def __init__(self):
        super(FrameViewer, self).__init__()
        self._empty = True

        # Initialize a flag which indicates when an image is being loaded.
        self.image_loading_busy = False

        # Initialize the photo object. No image is loaded yet.
        self._photo = QtWidgets.QGraphicsPixmapItem()
        self.shape_y = None
        self.shape_x = None

        self.viewrect = self.viewport().rect()

        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setResizeAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setBackgroundBrush(QtGui.QBrush(QtGui.QColor(30, 30, 30)))
        self.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.setDragMode(QtWidgets.QGraphicsView.ScrollHandDrag)
        self.drag_mode = True

        # Set the focus on the viewer.
        self.setFocus()

    def resizeEvent(self, event):
        """
        This method is called when the window size changes. In this case the image is zoomed so it
        fills the entire view.

        :param event: Resize event.
        :return: -
        """

        # Set the rectangle surrounding the current view.
        self.viewrect = self.viewport().rect()
        self.fitInView()
        return super(FrameViewer, self).resizeEvent(event)

    def hasPhoto(self):
        return not self._empty

    def get_zoom_factor(self):
        """
        Compute the current zoom factor. It is sent to the GUI widget via the "zoom_factor_signal".

        :return: The current zoom factor, rounded to the next int.
        """

        zoom_factor = round(self.scenerect.width() / self.photorect.width() * 100.)
        self.zoom_factor_signal.emit(zoom_factor)
        return zoom_factor

    def fitInView(self):
        """
        Scale the scene such that it fits into the window completely.

        :return: -
        """

        if not self.photorect.isNull():
            if self.hasPhoto():
                factor = min(self.viewrect.width() / self.scenerect.width(),
                             self.viewrect.height() / self.scenerect.height())
                self.scale(factor, factor)
                self.scenerect = self.transform().mapRect(self.photorect)
                self.get_zoom_factor()

    def set_original_scale(self):
        """
        Scale the scene to the original size of the photo. If the photo has more pixels than the
        current view, only part of the image is displayed.

        :return: -
        """

        if not self.photorect.isNull():
            if self.hasPhoto():
                factor = min(self.photorect.width() / self.scenerect.width(),
                             self.photorect.height() / self.scenerect.height())
                self.scale(factor, factor)
                self.scenerect = self.transform().mapRect(self.photorect)
                self.get_zoom_factor()

    def setPhoto(self, image, overlay_exclude_mark=False):
        """
        Convert a color or grayscale image to a pixmap and assign it to the photo object.

        :param image: Image to be displayed. The image is assumed to be in color or grayscale
                      format of length uint8 or uint16.
        :param overlay_exclude_mark: If True, overlay a crossed-out red circle in the upper left
                                     corner of the image.
        :return: -
        """

        # Indicate that an image is being loaded.
        self.image_loading_busy = True

        # Convert the image into uint8 format. If the frame type is uint16, values correspond to
        # 16bit resolution.
        if image.dtype == uint16:
            image_uint8 = (image / 256.).astype(uint8)
        elif image.dtype == uint8:
            image_uint8 = image.astype(uint8)
        elif image.dtype == float64:
            image_uint8 = image.astype(uint8)
        else:
            raise NotSupportedError("Attempt to set a photo in frame viewer with type neither"
                                    " uint8 nor uint16 nor float64")

        self.shape_y = image_uint8.shape[0]
        self.shape_x = image_uint8.shape[1]

        # Normalize the frame brightness.
        image_uint8 = normalize(image_uint8, None, alpha=0, beta=255, norm_type=NORM_MINMAX)

        # Overlay a crossed-out red circle in the upper left image corner.
        if overlay_exclude_mark:
            if len(image_uint8.shape) == 2:
                image_uint8 = cvtColor(image_uint8, COLOR_GRAY2RGB)

            # The position and size of the mark are relative to the image size.
            pos_x = int(round(self.shape_x / 10))
            pos_y = int(round(self.shape_y / 10))
            diameter = int(round(min(pos_x, pos_y) / 4))
            circle(image_uint8, (pos_x, pos_y), diameter, (255, 0, 0), 2)
            line(image_uint8, (pos_x - diameter, pos_y + diameter),
                 (pos_x + diameter, pos_y - diameter), (255, 0, 0), 2)

        # The image is monochrome:
        if len(image_uint8.shape) == 2:
            qt_image = QtGui.QImage(image_uint8, self.shape_x, self.shape_y, self.shape_x,
                                    QtGui.QImage.Format_Grayscale8)
        # The image is RGB color.
        else:
            qt_image = QtGui.QImage(image_uint8, self.shape_x,
                                    self.shape_y, 3 * self.shape_x, QtGui.QImage.Format_RGB888)
        pixmap = QtGui.QPixmap(qt_image)

        if pixmap and not pixmap.isNull():
            self._empty = False
            self._photo.setPixmap(pixmap)
        else:
            self._empty = True
            self._photo.setPixmap(QtGui.QPixmap())

        # Initialize the scene. This object handles mouse events if not in drag mode.
        self.initialize_scene()

        # Set the new rectangles surrounding the photo and the current scene.
        self.photorect = QtCore.QRectF(self._photo.pixmap().rect())
        self.scenerect = self.transform().mapRect(self.photorect)

        # Release the image loading flag.
        self.image_loading_busy = False

    def initialize_scene(self):
        """
        Initialize the scene. This object handles mouse events if not in drag mode. In derived
        viewer classes this method is replaced with the instantiation of a custom version of the
        graphics scene.

        :return:
        """

        self._scene = QtWidgets.QGraphicsScene()
        self._scene.addItem(self._photo)
        self.setScene(self._scene)

    def wheelEvent(self, event):
        """
        Handle scroll events for zooming in and out of the scene. This is only active when a photo
        is loaded.

        :param event: wheel event object
        :return: -
        """

        if self.drag_mode:
            # Depending of wheel direction, set the direction value to greater or smaller than 1.
            self.zoom(event.angleDelta().y())

        # If not in drag mode, the wheel event is handled at the scene level.
        else:
            try:
                self._scene.wheelEvent(event)

            # This exception occurs if the user uses "Ctrl + Wheel action" in the postproc_editor.
            # There this activity does not make any sense, so just ignore the event.
            except TypeError:
                pass

    def zoom(self, direction):
        """
        Zoom in or out. This is only active when a photo is loaded. For small images (smaller than
        the current view), zooming out is limited by the original photo resolution. For larger
        images zooming out stops when the scene fills the entire view.

        :param direction: If > 0, zoom in, otherwise zoom out.
        :return: -
        """

        if self.hasPhoto():
            # Depending of direction value, set the zoom factor to greater or smaller than 1.
            if direction > 0:
                factor = 1.25
            else:
                min_factor = min(
                    min(self.photorect.width(), self.viewrect.width()) / self.scenerect.width(),
                    min(self.photorect.height(), self.viewrect.height()) / self.scenerect.height())
                factor = max(0.8, min_factor)

            self.scale(factor, factor)
            self.scenerect = self.transform().mapRect(self.photorect)
            self.get_zoom_factor()

    def keyPressEvent(self, event):
        """
        The + and - keys are used for zooming.

        :param event: event object
        :return: -
        """

        # If the "+" key is pressed, zoom in. If "-" is pressed, zoom out.
        if event.key() == QtCore.Qt.Key_Control:
            self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
            self.drag_mode = False
        elif event.key() == QtCore.Qt.Key_Plus and not event.modifiers() & QtCore.Qt.ControlModifier:
            self.zoom(1)
        elif event.key() == QtCore.Qt.Key_Minus and not event.modifiers() & QtCore.Qt.ControlModifier:
            self.zoom(-1)
        elif event.key() == QtCore.Qt.Key_1 and not event.modifiers() & QtCore.Qt.ControlModifier:
            self.set_original_scale()
        else:
            super(FrameViewer, self).keyPressEvent(event)

    def keyReleaseEvent(self, event):
        # If the control key is released, switch back to "drag mode".
        # Use default handling for other keys.
        if event.key() == QtCore.Qt.Key_Control:
            self.setDragMode(QtWidgets.QGraphicsView.ScrollHandDrag)
            self.drag_mode = True
        else:
            super(FrameViewer, self).keyReleaseEvent(event)


class VideoFrameViewer(FrameViewer):
    """
    This widget implements a frame viewer for frames in a video file. Panning and zooming is
    implemented by using the mouse and scroll wheel.

    """

    def __init__(self, frames, align_frames, frame_index=0):
        super(VideoFrameViewer, self).__init__()
        self.frames = frames
        self.align_frames = align_frames
        self.frame_index = frame_index

        self.setPhoto(self.frame_index)

    def setPhoto(self, index):
        """
        Convert a grayscale image to a pixmap and assign it to the photo object.

        :param index: Index into the frame list. Frames are assumed to be grayscale image in format
                      uint8 or uint16.
        :return: -
        """

        # Indicate that an image is being loaded.
        self.image_loading_busy = True

        image = self.frames.frames_mono(index)[self.align_frames.intersection_shape[0][0] -
                                                    self.align_frames.frame_shifts[index][0]:
                                                    self.align_frames.intersection_shape[0][1] -
                                                    self.align_frames.frame_shifts[index][0],
                                                    self.align_frames.intersection_shape[1][0] -
                                                    self.align_frames.frame_shifts[index][1]:
                                                    self.align_frames.intersection_shape[1][1] -
                                                    self.align_frames.frame_shifts[index][1]]

        super(VideoFrameViewer, self).setPhoto(image)


class FrameViewerWidget(QtWidgets.QFrame, Ui_frame_viewer):
    """
    This widget implements a frame viewer together with control elements to visualize frame
    qualities, and to manipulate the stack limits.
    """

    frame_player_start_signal = QtCore.pyqtSignal()

    def __init__(self, parent_gui, configuration, frames, rank_frames, align_frames,
                 stacked_image_log_file, signal_finished, signal_payload):
        """
        Initialization of the widget.

        :param parent_gui: Parent GUI object
        :param configuration: Configuration object with parameters
        :param frames: Frames object with all video frames
        :param rank_frames: RankFrames object with global quality ranks (between 0. and 1.,
                            1. being optimal) for all frames
        :param align_frames: AlignFrames object with global shift information for all frames
        :param stacked_image_log_file: Log file to be stored with results, or None.
        :param signal_finished: Qt signal with signature (str) to trigger the next activity when
                                the viewer exits.
        :param signal_payload: Payload of "signal_finished" (str).
        """

        super(FrameViewerWidget, self).__init__(parent_gui)
        self.setupUi(self)

        # Keep references to upper level objects.
        self.parent_gui = parent_gui
        self.configuration = configuration
        self.stacked_image_log_file = stacked_image_log_file
        self.signal_finished = signal_finished
        self.signal_payload = signal_payload
        self.frames = frames
        self.rank_frames = rank_frames
        self.align_frames = align_frames

        # Be careful: Indices are counted from 0, while widget contents are counted from 1 (to make
        # it easier for the user.
        self.quality_index = 0
        self.frame_index = self.rank_frames.quality_sorted_indices[self.quality_index]

        # Set up the frame viewer and put it in the upper left corner.
        self.frame_viewer = VideoFrameViewer(self.frames, self.align_frames, self.frame_index)
        self.frame_viewer.setObjectName("framewiever")
        self.grid_layout.addWidget(self.frame_viewer, 0, 0, 4, 3)

        self.widget_elements = [self.spinBox_chronological,
                                self.spinBox_quality,
                                self.slider_frames,
                                self.pushButton_play]

        # Initialize variables. The values for "alignment_point_frame_number" and
        # "alignment_points_frame_percent" are held as copies in this object. Only if the user
        # presses "OK" at the end, the values are copied back into the configuration object.
        self.stack_size_changed = None

        # The number of frames to be stacked is selected explicitly.
        if self.configuration.alignment_points_frame_number > 0:
            self.alignment_points_frame_number = min(
                self.configuration.alignment_points_frame_number, self.frames.number)
            self.alignment_points_frame_percent = max(1, min(round(
                self.alignment_points_frame_number / self.frames.number * 100.), 100))
        else:
            self.alignment_points_frame_percent = self.configuration.alignment_points_frame_percent
            self.alignment_points_frame_number = max(1, round(
                self.frames.number * self.alignment_points_frame_percent / 100.))

        self.frame_ranks = rank_frames.frame_ranks
        self.quality_sorted_indices = rank_frames.quality_sorted_indices

        # Start with ordering frames by quality. This can be changed by the user using a radio
        # button.
        self.frame_ordering = "quality"

        # Initialize a variable for communication with the frame_player object later.
        self.run_player = False

        # Create the frame player thread and start it. The player displays frames in succession.
        # It is pushed on a different thread because otherwise the user could not stop it before it
        # finishes.
        self.frame_player = FramePlayer(self)
        self.player_thread = QtCore.QThread()
        self.frame_player.setParent(None)
        self.frame_player.moveToThread(self.player_thread)
        self.frame_player.block_widgets_signal.connect(self.block_widgets)
        self.frame_player.unblock_widgets_signal.connect(self.unblock_widgets)
        self.frame_player.set_photo_signal.connect(self.frame_viewer.setPhoto)
        self.frame_player.set_slider_value.connect(self.slider_frames.setValue)
        self.frame_player_start_signal.connect(self.frame_player.play)
        self.player_thread.start()

        # Initialization of GUI elements
        self.slider_frames.setMinimum(1)
        self.slider_frames.setMaximum(self.frames.number)
        self.slider_frames.setValue(self.quality_index + 1)
        self.spinBox_chronological.setMinimum(1)
        self.spinBox_chronological.setMaximum(self.frames.number)
        self.spinBox_chronological.setValue(self.frame_index + 1)
        self.spinBox_quality.setMinimum(1)
        self.spinBox_quality.setMaximum(self.frames.number)
        self.spinBox_quality.setValue(self.quality_index + 1)
        self.radioButton_quality.setChecked(True)

        self.spinBox_number_frames.setMaximum(self.frames.number)
        self.spinBox_number_frames.setValue(self.alignment_points_frame_number)
        self.spinBox_percentage_frames.setValue(self.alignment_points_frame_percent)

        # Create the Matplotlib widget showing the quality lines.
        self.matplotlib_widget = MatplotlibWidget(self.configuration, self.rank_frames)

        self.grid_layout.addWidget(Canvas(self.matplotlib_widget.fig), 0, 3, 2, 1)

        self.grid_layout.setColumnStretch(0, 5)
        self.grid_layout.setColumnStretch(1, 0)
        self.grid_layout.setColumnStretch(2, 0)
        self.grid_layout.setColumnStretch(3, 1)
        self.grid_layout.setRowStretch(0, 1)
        self.grid_layout.setRowStretch(1, 0)
        self.grid_layout.setRowStretch(2, 0)
        self.grid_layout.setRowStretch(3, 0)

        # Connect signals with slots.
        self.buttonBox.accepted.connect(self.done)
        self.buttonBox.rejected.connect(self.reject)
        self.slider_frames.valueChanged.connect(self.slider_frames_changed)
        self.spinBox_number_frames.valueChanged.connect(self.spinbox_number_frames_changed)
        self.spinBox_percentage_frames.valueChanged.connect(self.spinbox_percentage_frames_changed)
        self.radioButton_quality.toggled.connect(self.radiobutton_quality_changed)
        self.spinBox_chronological.valueChanged.connect(self.spinbox_chronological_changed)
        self.spinBox_quality.valueChanged.connect(self.spinbox_quality_changed)
        self.pushButton_set_stacking_limit.clicked.connect(
            self.pushbutton_set_stacking_limit_clicked)
        self.pushButton_play.clicked.connect(self.pushbutton_play_clicked)
        self.pushButton_stop.clicked.connect(self.pushbutton_stop_clicked)

        # Initialize the Matplotlib widget contents.
        self.matplotlib_widget.renew_plot(self.quality_index, self.frame_ordering,
                                          self.alignment_points_frame_number)

    @QtCore.pyqtSlot()
    def block_widgets(self):
        """
        Disable GUI elements to prevent unwanted user interaction while the frame player is running.

        :return: -
        """

        for element in self.widget_elements:
            element.setDisabled(True)

    @QtCore.pyqtSlot()
    def unblock_widgets(self):
        """
        Enable GUI elements which were temporarily blocked when the player stops.

        :return: -
        """

        for element in self.widget_elements:
            element.setDisabled(False)

    def slider_frames_changed(self):
        """
        The frames slider is changed by the user.

        :return: -
        """

        # Again, please note the difference between indexing and GUI displays.
        index = self.slider_frames.value() - 1

        # Differentiate between frame ordering (by quality or chronologically).
        if self.frame_ordering == "quality":
            self.frame_index = self.rank_frames.quality_sorted_indices[index]
            self.quality_index = index

            # Plot a dot on the quality line at the position of the current frame.
            self.matplotlib_widget.plot_dot(self.quality_index)
        else:
            self.frame_index = index
            self.quality_index = self.rank_frames.quality_sorted_indices.index(self.frame_index)
            self.matplotlib_widget.plot_dot(self.frame_index)

        # Block signals temporarily to avoid feedback loops. Then update widget contents.
        self.spinBox_chronological.blockSignals(True)
        self.spinBox_quality.blockSignals(True)
        self.spinBox_chronological.setValue(self.frame_index + 1)
        self.spinBox_quality.setValue(self.quality_index + 1)
        self.spinBox_chronological.blockSignals(False)
        self.spinBox_quality.blockSignals(False)
        self.frame_viewer.setPhoto(self.frame_index)

    def spinbox_number_frames_changed(self):
        """
        The user has changed the number of frames to be stacked at each AP.

        :return: -
        """

        self.alignment_points_frame_number = self.spinBox_number_frames.value()
        self.alignment_points_frame_percent = int(
            round(self.alignment_points_frame_number * 100. / self.frames.number))
        self.spinBox_percentage_frames.blockSignals(True)
        self.spinBox_percentage_frames.setValue(self.alignment_points_frame_percent)
        self.spinBox_percentage_frames.blockSignals(False)
        self.matplotlib_widget.plot_cutoff_lines(self.frame_ordering,
                                                 self.alignment_points_frame_number)
        self.stack_size_changed = 'number'

    def spinbox_percentage_frames_changed(self):
        """
        The user has changed the percentage of frames to be stacked at each AP.

        :return:
        """

        self.alignment_points_frame_percent = self.spinBox_percentage_frames.value()
        self.alignment_points_frame_number = max(1, round(
            self.frames.number * self.alignment_points_frame_percent / 100.))
        self.spinBox_number_frames.blockSignals(True)
        self.spinBox_number_frames.setValue(self.alignment_points_frame_number)
        self.spinBox_number_frames.blockSignals(False)
        self.matplotlib_widget.plot_cutoff_lines(self.frame_ordering,
                                                 self.alignment_points_frame_number)
        self.stack_size_changed = 'percent'

    def radiobutton_quality_changed(self):
        """
        Toggle back and forth between frame ordering modes. The frame slider is updated to reflect
        the index of the current frame in the new ordering scheme. The Matplotlib widget changes
        its appearance.

        :return: -
        """

        if self.frame_ordering == "quality":
            self.frame_ordering = "chronological"
            self.slider_frames.setValue(self.frame_index + 1)
            self.matplotlib_widget.renew_plot(self.frame_index, self.frame_ordering,
                                              self.alignment_points_frame_number)
        else:
            self.frame_ordering = "quality"
            self.slider_frames.setValue(self.quality_index + 1)
            self.matplotlib_widget.renew_plot(self.quality_index, self.frame_ordering,
                                              self.alignment_points_frame_number)

    def spinbox_chronological_changed(self):
        """
        The user has selected a new frame to be displayed by entering its chronological frame index.
        All other widgets which depend on the current frame index are changed.

        :return: -
        """

        self.frame_index = self.spinBox_chronological.value() - 1
        self.quality_index = self.rank_frames.quality_sorted_indices.index(self.frame_index)
        self.slider_frames.blockSignals(True)
        self.spinBox_quality.blockSignals(True)
        self.spinBox_quality.setValue(self.quality_index + 1)
        if self.frame_ordering == "quality":
            self.slider_frames.setValue(self.quality_index + 1)
            self.matplotlib_widget.plot_dot(self.quality_index)
        else:
            self.slider_frames.setValue(self.frame_index + 1)
            self.matplotlib_widget.plot_dot(self.frame_index)
        self.slider_frames.blockSignals(False)
        self.spinBox_quality.blockSignals(False)
        self.frame_viewer.setPhoto(self.frame_index)

    def spinbox_quality_changed(self):
        """
        The user has selected a new frame to be displayed by entering its quality index.
        All other widgets which depend on the current frame index are changed.

        :return:
        """

        self.quality_index = self.spinBox_quality.value() - 1
        self.frame_index = self.rank_frames.quality_sorted_indices[self.quality_index]
        self.slider_frames.blockSignals(True)
        self.spinBox_chronological.blockSignals(True)
        self.spinBox_chronological.setValue(self.frame_index + 1)
        if self.frame_ordering == "quality":
            self.slider_frames.setValue(self.quality_index + 1)
            self.matplotlib_widget.plot_dot(self.quality_index)
        else:
            self.slider_frames.setValue(self.frame_index + 1)
            self.matplotlib_widget.plot_dot(self.frame_index)
        self.slider_frames.blockSignals(False)
        self.spinBox_chronological.blockSignals(False)
        self.frame_viewer.setPhoto(self.frame_index)

    def pushbutton_set_stacking_limit_clicked(self):
        """
        The current frame defines the stacking limit. Compute the corresponding stack limit
        variables and update the widgets. Since the signals are not deactivated before updating
        the widgets, the Matplotlib widget gets updated as well.
        :return:
        """
        self.alignment_points_frame_number = self.quality_index + 1
        self.alignment_points_frame_percent = max(1, round(
            self.alignment_points_frame_number * 100. / self.frames.number))
        self.spinBox_number_frames.setValue(self.alignment_points_frame_number)
        self.spinBox_percentage_frames.setValue(self.alignment_points_frame_percent)
        self.stack_size_changed = 'number'

    @QtCore.pyqtSlot()
    def pushbutton_play_clicked(self):
        """
        Start the frame player if it is not running already.

        :return: -
        """

        self.frame_player_start_signal.emit()

    def pushbutton_stop_clicked(self):
        """
        When the frame player is running, it periodically checks this variable. If it is set to
        False, the player stops.

        :return:
        """

        self.frame_player.run_player = False

    def done(self):
        """
        On exit from the frame viewer, update the stack frame size and send a completion signal.

        :return: -
        """

        # Check if a new stack size was selected.
        if self.stack_size_changed is not None:

            # The number of frames has been set explicitly. Check if it has changed.
            if self.stack_size_changed == 'number':
                if self.spinBox_number_frames.value() != self.configuration.alignment_points_frame_number:
                    self.configuration.alignment_points_frame_number = self.spinBox_number_frames.value()
                    self.configuration.alignment_points_frame_percent = -1
                    if self.configuration.global_parameters_protocol_level > 1:
                        Miscellaneous.protocol(
                            "           The user has selected a new stack size: " + str(
                                self.configuration.alignment_points_frame_number) + " frames.",
                            self.stacked_image_log_file, precede_with_timestamp=False)

            # The percentage of frames has been set. Check if it has changed.
            elif self.spinBox_percentage_frames.value() != self.configuration.alignment_points_frame_percent:
                self.configuration.alignment_points_frame_number = -1
                self.configuration.alignment_points_frame_percent = \
                    self.spinBox_percentage_frames.value()
                if self.configuration.global_parameters_protocol_level > 1:
                    Miscellaneous.protocol(
                        "           The user has selected a new stack size: " + str(
                            self.configuration.alignment_points_frame_percent) + "% of all frames.",
                        self.stacked_image_log_file, precede_with_timestamp=False)

        # Send a completion message.
        if self.parent_gui is not None:
            self.signal_finished.emit(self.signal_payload)

        # Close the Window.
        self.player_thread.quit()
        plt.close()
        self.close()

    def reject(self):
        """
        If the "cancel" button is pressed, the coordinates are not stored.

        :return: -
        """

        # Send a completion message.
        if self.parent_gui is not None:
            self.signal_finished.emit(self.signal_payload)

        # Close the Window.
        self.player_thread.quit()
        plt.close()
        self.close()


class FramePlayer(QtCore.QObject):
    """
    This class implements a video player using the FrameViewer and the control elements of the
    FrameViewerWidget. The player is started by the widget on a separate thread. This way the user
    can instruct the GUI to stop the running player.

    """

    block_widgets_signal = QtCore.pyqtSignal()
    unblock_widgets_signal = QtCore.pyqtSignal()
    set_photo_signal = QtCore.pyqtSignal(int)
    set_slider_value = QtCore.pyqtSignal(int)

    def __init__(self, frame_viewer_widget, parent=None):
        super(FramePlayer, self).__init__()

        # Store a reference of the frame viewer widget.
        self.frame_viewer_widget = frame_viewer_widget

        # Set the delay time between frames.
        self.delay_between_frames = 0.1

        # Initialize a variable used to stop the player in the GUI thread.
        self.run_player = False

    @QtCore.pyqtSlot()
    def play(self):
        """
        Start the player.
        :return: -
        """

        # Block signals from GUI elements to avoid cross-talk, and disable them to prevent unwanted
        # user interaction.
        self.block_widgets_signal.emit()

        # Set the plaer running.
        self.run_player = True

        # The frames are ordered by their quality.
        if self.frame_viewer_widget.frame_ordering == "quality":

            # The player stops when the end of the video is reached, or when the "run_player"
            # variable is set to False in the GUI thread.
            while self.frame_viewer_widget.quality_index < self.frame_viewer_widget.frames.number \
                    - 1 and self.run_player:
                # Go to the next frame only if the viewer has finished loading the previous one.
                if not self.frame_viewer_widget.frame_viewer.image_loading_busy:
                    self.frame_viewer_widget.quality_index += 1
                    self.frame_viewer_widget.frame_index = \
                    self.frame_viewer_widget.rank_frames.quality_sorted_indices[
                        self.frame_viewer_widget.quality_index]
                    self.set_slider_value.emit(self.frame_viewer_widget.quality_index + 1)
                    self.set_photo_signal.emit(self.frame_viewer_widget.frame_index)

                # Insert a short pause to keep the video from running too fast.
                sleep(self.delay_between_frames)

        else:
            # The same for chronological frame ordering.
            while self.frame_viewer_widget.frame_index < self.frame_viewer_widget.frames.number \
                    - 1 and self.run_player:
                if not self.frame_viewer_widget.frame_viewer.image_loading_busy:
                    self.frame_viewer_widget.frame_index += 1
                    self.frame_viewer_widget.quality_index = \
                        self.frame_viewer_widget.rank_frames.quality_sorted_indices.index(
                        self.frame_viewer_widget.frame_index)
                    self.set_slider_value.emit(self.frame_viewer_widget.frame_index + 1)
                    self.set_photo_signal.emit(self.frame_viewer_widget.frame_index)
                sleep(self.delay_between_frames)

        self.run_player = False

        # Re-set the GUI elements to their normal state.
        self.unblock_widgets_signal.emit()


if __name__ == '__main__':
    # Images can either be extracted from a video file or a batch of single photographs. Select
    # the example for the test run.
    type = 'video'
    if type == 'image':
        names = glob(
            'Images/2012*.tif')  # names = glob.glob('Images/Moon_Tile-031*ap85_8b.tif')  # names
        # = glob.glob('Images/Example-3*.jpg')
    else:
        names = 'Videos/another_short_video.avi'
    print(names)

    # Get configuration parameters.
    configuration = Configuration()
    configuration.initialize_configuration()
    try:
        frames = Frames(configuration, names, type=type)
        print("Number of images read: " + str(frames.number))
        print("Image shape: " + str(frames.shape))
    except Error as e:
        print("Error: " + e.message)
        exit()

    # Rank the frames by their overall local contrast.
    rank_frames = RankFrames(frames, configuration)
    start = time()
    rank_frames.frame_score()
    end = time()
    print('Elapsed time in ranking images: {}'.format(end - start))
    print("Index of maximum: " + str(rank_frames.frame_ranks_max_index))
    print("Frame scores: " + str(rank_frames.frame_ranks))
    print("Frame scores (sorted): " + str(
        [rank_frames.frame_ranks[i] for i in rank_frames.quality_sorted_indices]))
    print("Sorted index list: " + str(rank_frames.quality_sorted_indices))

    # Initialize the frame alignment object.
    align_frames = AlignFrames(frames, rank_frames, configuration)

    if configuration.align_frames_mode == "Surface":
        # Select the local rectangular patch in the image where the L gradient is highest in both x
        # and y direction. The scale factor specifies how much smaller the patch is compared to the
        # whole image frame.
        (y_low_opt, y_high_opt, x_low_opt, x_high_opt) = align_frames.compute_alignment_rect(
            configuration.align_frames_rectangle_scale_factor)

        print("optimal alignment rectangle, y_low: " + str(y_low_opt) + ", y_high: " +
              str(y_high_opt) + ", x_low: " + str(x_low_opt) + ", x_high: " + str(x_high_opt))

    # Align all frames globally relative to the frame with the highest score.
    try:
        align_frames.align_frames()
    except NotSupportedError as e:
        print("Error: " + e.message)
        exit()
    except InternalError as e:
        print("Warning: " + e.message)

    print("Intersection, y_low: " + str(align_frames.intersection_shape[0][0]) + ", y_high: "
          + str(align_frames.intersection_shape[0][1]) + ", x_low: " \
          + str(align_frames.intersection_shape[1][0]) + ", x_high: " \
          + str(align_frames.intersection_shape[1][1]))

    app = QtWidgets.QApplication(argv)
    window = FrameViewerWidget(None, configuration, frames, rank_frames, align_frames, None, None,
                               None)
    window.setMinimumSize(800, 600)
    window.showMaximized()
    app.exec_()

    print("Percentage of frames to be stacked: " + str(
        configuration.alignment_points_frame_percent) + ", number of frames: " + str(
        configuration.alignment_points_frame_number))

    exit()



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 25 19:54:17 2019
@author: Jens.Scheidtmann@gmail.com
Run
> python -m cProfiler -o profile_info.txt main_program.py
to create profiling information, then run this file to output the top 10 function calls
(by cumulative time).
Make sure that you have set 'show_results = False' in the 'Specify Test Case'
section in main_program.py (else some graphic displays are opened, which
wait for the user to close the window)
"""


import pstats
from pstats import SortKey
p = pstats.Stats('profile_info.txt')
p.sort_stats(SortKey.CUMULATIVE).print_stats(10)

# p.strip_dirs().sort_stats(-1).print_stats()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rectangular_patch_editor_gui.ui'
#
# Created by: PyQt5 UI code generator 5.12
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_rectangular_patch_editor(object):
    def setupUi(self, rectangular_patch_editor):
        rectangular_patch_editor.setObjectName("rectangular_patch_editor")
        rectangular_patch_editor.resize(900, 630)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(rectangular_patch_editor.sizePolicy().hasHeightForWidth())
        rectangular_patch_editor.setSizePolicy(sizePolicy)
        rectangular_patch_editor.setLocale(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.UnitedStates))
        rectangular_patch_editor.setFrameShape(QtWidgets.QFrame.Panel)
        rectangular_patch_editor.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.verticalLayout = QtWidgets.QVBoxLayout(rectangular_patch_editor)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.messageLabel = QtWidgets.QLabel(rectangular_patch_editor)
        self.messageLabel.setObjectName("messageLabel")
        self.horizontalLayout.addWidget(self.messageLabel)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.buttonBox = QtWidgets.QDialogButtonBox(rectangular_patch_editor)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.horizontalLayout.addWidget(self.buttonBox)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(rectangular_patch_editor)
        QtCore.QMetaObject.connectSlotsByName(rectangular_patch_editor)

    def retranslateUi(self, rectangular_patch_editor):
        _translate = QtCore.QCoreApplication.translate
        rectangular_patch_editor.setWindowTitle(_translate("rectangular_patch_editor", "Frame"))
        self.messageLabel.setText(_translate("rectangular_patch_editor", "Message to be displayed"))
        self.buttonBox.setToolTip(_translate("rectangular_patch_editor", "Press \'OK\' to save the selection and exit, or \'cancel\' to discard the selection."))





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'sharpening_layer_widget.ui'
#
# Created by: PyQt5 UI code generator 5.15.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_sharpening_layer_widget(object):
    def setupUi(self, sharpening_layer_widget):
        sharpening_layer_widget.setObjectName("sharpening_layer_widget")
        sharpening_layer_widget.resize(311, 227)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(sharpening_layer_widget.sizePolicy().hasHeightForWidth())
        sharpening_layer_widget.setSizePolicy(sizePolicy)
        self.verticalLayout = QtWidgets.QVBoxLayout(sharpening_layer_widget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.groupBox_layer = QtWidgets.QGroupBox(sharpening_layer_widget)
        self.groupBox_layer.setObjectName("groupBox_layer")
        self.gridLayout = QtWidgets.QGridLayout(self.groupBox_layer)
        self.gridLayout.setHorizontalSpacing(12)
        self.gridLayout.setObjectName("gridLayout")
        self.label_amount = QtWidgets.QLabel(self.groupBox_layer)
        self.label_amount.setObjectName("label_amount")
        self.gridLayout.addWidget(self.label_amount, 1, 1, 1, 1)
        self.checkBox_luminance = QtWidgets.QCheckBox(self.groupBox_layer)
        self.checkBox_luminance.setObjectName("checkBox_luminance")
        self.gridLayout.addWidget(self.checkBox_luminance, 7, 1, 1, 2)
        self.label_radius = QtWidgets.QLabel(self.groupBox_layer)
        self.label_radius.setObjectName("label_radius")
        self.gridLayout.addWidget(self.label_radius, 0, 1, 1, 1)
        self.pushButton_remove = QtWidgets.QPushButton(self.groupBox_layer)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_remove.sizePolicy().hasHeightForWidth())
        self.pushButton_remove.setSizePolicy(sizePolicy)
        self.pushButton_remove.setMinimumSize(QtCore.QSize(60, 0))
        self.pushButton_remove.setMaximumSize(QtCore.QSize(60, 16777215))
        self.pushButton_remove.setObjectName("pushButton_remove")
        self.gridLayout.addWidget(self.pushButton_remove, 7, 3, 1, 1)
        self.lineEdit_radius = QtWidgets.QLineEdit(self.groupBox_layer)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEdit_radius.sizePolicy().hasHeightForWidth())
        self.lineEdit_radius.setSizePolicy(sizePolicy)
        self.lineEdit_radius.setMinimumSize(QtCore.QSize(60, 0))
        self.lineEdit_radius.setMaximumSize(QtCore.QSize(60, 16777215))
        self.lineEdit_radius.setObjectName("lineEdit_radius")
        self.gridLayout.addWidget(self.lineEdit_radius, 0, 3, 1, 1)
        self.horizontalSlider_bi_fraction = QtWidgets.QSlider(self.groupBox_layer)
        self.horizontalSlider_bi_fraction.setMaximum(100)
        self.horizontalSlider_bi_fraction.setPageStep(5)
        self.horizontalSlider_bi_fraction.setOrientation(QtCore.Qt.Horizontal)
        self.horizontalSlider_bi_fraction.setObjectName("horizontalSlider_bi_fraction")
        self.gridLayout.addWidget(self.horizontalSlider_bi_fraction, 3, 2, 1, 1)
        self.bilateral_fraction_label = QtWidgets.QLabel(self.groupBox_layer)
        self.bilateral_fraction_label.setObjectName("bilateral_fraction_label")
        self.gridLayout.addWidget(self.bilateral_fraction_label, 3, 1, 1, 1)
        self.lineEdit_bi_fraction = QtWidgets.QLineEdit(self.groupBox_layer)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEdit_bi_fraction.sizePolicy().hasHeightForWidth())
        self.lineEdit_bi_fraction.setSizePolicy(sizePolicy)
        self.lineEdit_bi_fraction.setMinimumSize(QtCore.QSize(60, 0))
        self.lineEdit_bi_fraction.setMaximumSize(QtCore.QSize(60, 16777215))
        self.lineEdit_bi_fraction.setObjectName("lineEdit_bi_fraction")
        self.gridLayout.addWidget(self.lineEdit_bi_fraction, 3, 3, 1, 1)
        self.lineEdit_amount = QtWidgets.QLineEdit(self.groupBox_layer)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEdit_amount.sizePolicy().hasHeightForWidth())
        self.lineEdit_amount.setSizePolicy(sizePolicy)
        self.lineEdit_amount.setMinimumSize(QtCore.QSize(60, 0))
        self.lineEdit_amount.setMaximumSize(QtCore.QSize(60, 16777215))
        self.lineEdit_amount.setObjectName("lineEdit_amount")
        self.gridLayout.addWidget(self.lineEdit_amount, 1, 3, 1, 1)
        self.horizontalSlider_amount = QtWidgets.QSlider(self.groupBox_layer)
        self.horizontalSlider_amount.setMaximum(100)
        self.horizontalSlider_amount.setPageStep(1)
        self.horizontalSlider_amount.setProperty("value", 3)
        self.horizontalSlider_amount.setOrientation(QtCore.Qt.Horizontal)
        self.horizontalSlider_amount.setObjectName("horizontalSlider_amount")
        self.gridLayout.addWidget(self.horizontalSlider_amount, 1, 2, 1, 1)
        self.horizontalSlider_radius = QtWidgets.QSlider(self.groupBox_layer)
        self.horizontalSlider_radius.setMinimum(1)
        self.horizontalSlider_radius.setMaximum(100)
        self.horizontalSlider_radius.setPageStep(1)
        self.horizontalSlider_radius.setOrientation(QtCore.Qt.Horizontal)
        self.horizontalSlider_radius.setObjectName("horizontalSlider_radius")
        self.gridLayout.addWidget(self.horizontalSlider_radius, 0, 2, 1, 1)
        self.label_bi_range = QtWidgets.QLabel(self.groupBox_layer)
        self.label_bi_range.setObjectName("label_bi_range")
        self.gridLayout.addWidget(self.label_bi_range, 4, 1, 1, 1)
        self.horizontalSlider_bi_range = QtWidgets.QSlider(self.groupBox_layer)
        self.horizontalSlider_bi_range.setMaximum(200)
        self.horizontalSlider_bi_range.setPageStep(2)
        self.horizontalSlider_bi_range.setOrientation(QtCore.Qt.Horizontal)
        self.horizontalSlider_bi_range.setObjectName("horizontalSlider_bi_range")
        self.gridLayout.addWidget(self.horizontalSlider_bi_range, 4, 2, 1, 1)
        self.lineEdit_bi_range = QtWidgets.QLineEdit(self.groupBox_layer)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEdit_bi_range.sizePolicy().hasHeightForWidth())
        self.lineEdit_bi_range.setSizePolicy(sizePolicy)
        self.lineEdit_bi_range.setMaximumSize(QtCore.QSize(60, 16777215))
        self.lineEdit_bi_range.setObjectName("lineEdit_bi_range")
        self.gridLayout.addWidget(self.lineEdit_bi_range, 4, 3, 1, 1)
        self.label_denoise = QtWidgets.QLabel(self.groupBox_layer)
        self.label_denoise.setObjectName("label_denoise")
        self.gridLayout.addWidget(self.label_denoise, 5, 1, 1, 1)
        self.horizontalSlider_denoise = QtWidgets.QSlider(self.groupBox_layer)
        self.horizontalSlider_denoise.setMaximum(100)
        self.horizontalSlider_denoise.setOrientation(QtCore.Qt.Horizontal)
        self.horizontalSlider_denoise.setObjectName("horizontalSlider_denoise")
        self.gridLayout.addWidget(self.horizontalSlider_denoise, 5, 2, 1, 1)
        self.lineEdit_denoise = QtWidgets.QLineEdit(self.groupBox_layer)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEdit_denoise.sizePolicy().hasHeightForWidth())
        self.lineEdit_denoise.setSizePolicy(sizePolicy)
        self.lineEdit_denoise.setMaximumSize(QtCore.QSize(60, 16777215))
        self.lineEdit_denoise.setObjectName("lineEdit_denoise")
        self.gridLayout.addWidget(self.lineEdit_denoise, 5, 3, 1, 1)
        self.verticalLayout.addWidget(self.groupBox_layer)

        self.retranslateUi(sharpening_layer_widget)
        QtCore.QMetaObject.connectSlotsByName(sharpening_layer_widget)

    def retranslateUi(self, sharpening_layer_widget):
        _translate = QtCore.QCoreApplication.translate
        sharpening_layer_widget.setWindowTitle(_translate("sharpening_layer_widget", "Form"))
        self.groupBox_layer.setTitle(_translate("sharpening_layer_widget", "Layer 1"))
        self.label_amount.setText(_translate("sharpening_layer_widget", "Amount"))
        self.checkBox_luminance.setToolTip(_translate("sharpening_layer_widget", "Apply sharpening only to luminance channel"))
        self.checkBox_luminance.setText(_translate("sharpening_layer_widget", "Luminance channel only"))
        self.label_radius.setText(_translate("sharpening_layer_widget", "Radius"))
        self.pushButton_remove.setToolTip(_translate("sharpening_layer_widget", "Remove this layer"))
        self.pushButton_remove.setText(_translate("sharpening_layer_widget", "Remove"))
        self.lineEdit_radius.setToolTip(_translate("sharpening_layer_widget", "Radius of Gauss filter"))
        self.horizontalSlider_bi_fraction.setToolTip(_translate("sharpening_layer_widget", "Set the fraction of the bilateral filter component (vs. Gaussian)"))
        self.bilateral_fraction_label.setText(_translate("sharpening_layer_widget", "<html><head/><body><p>Bilateral<br/>Fraction</p></body></html>"))
        self.lineEdit_bi_fraction.setToolTip(_translate("sharpening_layer_widget", "Fraction of bilateral filter component"))
        self.lineEdit_amount.setToolTip(_translate("sharpening_layer_widget", "Relative effect of this layer"))
        self.horizontalSlider_amount.setToolTip(_translate("sharpening_layer_widget", "Increase / decrease the effect of this layer"))
        self.horizontalSlider_radius.setToolTip(_translate("sharpening_layer_widget", "Change the radius of Gauss filter for this layer"))
        self.label_bi_range.setText(_translate("sharpening_layer_widget", "<html><head/><body><p>Bilateral<br/>Range</p></body></html>"))
        self.horizontalSlider_bi_range.setToolTip(_translate("sharpening_layer_widget", "Set the brigthness range parameter of the bilateral filter"))
        self.lineEdit_bi_range.setToolTip(_translate("sharpening_layer_widget", "Range parameter of bilateral filter"))
        self.label_denoise.setText(_translate("sharpening_layer_widget", "Denoise"))
        self.horizontalSlider_denoise.setToolTip(_translate("sharpening_layer_widget", "Set the strength of the denoise filter (Gaussian) for this layer"))
        self.lineEdit_denoise.setToolTip(_translate("sharpening_layer_widget", "Strength of the denoise filter (Gaussian) for this layer"))



import cv2
from os import remove
from pathlib import Path
from cv2 import resize, INTER_CUBIC


def process_video(input_filename, output_filename, max_frames, frames_stride, reduce_frame_size):
    # Open the input video and read metadata.
    cap = cv2.VideoCapture(input_filename)
    frame_width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    frame_height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    fps = int(cap.get(cv2.CAP_PROP_FPS))
    # fourcc = cap.get(cv2.CAP_PROP_FOURCC)
    frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    convert_rgb = int(cap.get(cv2.CAP_PROP_CONVERT_RGB))
    print("Input video:\nFrame width: " + str(frame_width) + ", frame height: " + str(frame_height))
    print("Fps: " + str(fps))
    print("Number of frames: " + str(frame_count))
    print("Convert to RGB: " + str(convert_rgb))

    # Define the codec and create VideoWriter object
    if Path(output_filename).is_file():
        remove(output_filename)

    # reduce frame size to multiples of reduce_frame_size.
    frame_width = frame_width - frame_width % reduce_frame_size
    frame_height = frame_height - frame_height % reduce_frame_size
    frame_width_output = int(frame_width / reduce_frame_size)
    frame_height_output = int(frame_height / reduce_frame_size)

    print("Shape of output frames: (" + str(frame_width_output) + ", " + str(
        frame_height_output) + ")")

    fourcc = cv2.VideoWriter_fourcc('D', 'I', 'V', 'X')
    out = cv2.VideoWriter(output_filename, fourcc, fps, (frame_width_output, frame_height_output))

    frame_index_read = 0
    frame_index_written = 0

    while cap.isOpened():
        ret, frame = cap.read()
        if not frame_index_read % frames_stride:
            output_frame = resize(frame, (frame_width_output, frame_height_output),
                                  interpolation=INTER_CUBIC)
            out.write(output_frame)
            frame_index_written += 1
            if frame_index_written == max_frames:
                break
        frame_index_read += 1
        if frame_index_read == frame_count:
            break

    cap.release()
    out.release()

    print("Number of frames read: " + str(frame_index_read))
    print("Number of frames written: " + str(frame_index_written))


# input_filename = 'D:\SW-Development\Python\PlanetarySystemStacker\Source\Videos\Moon_Tile-013_205538.avi'
input_filename = 'D:\SW-Development\Python\PlanetarySystemStacker\Examples\Moon_2018-03-24\Moon_Tile-024_043939.avi'
output_filename = 'D:\SW-Development\Python\PlanetarySystemStacker\Examples\Moon_2018-03-24\Moon_Tile-024_043939_resampled.avi'
max_frames = 1000
frames_stride = 20
reduce_frame_size = 4

process_video(input_filename, output_filename, max_frames, frames_stride, reduce_frame_size)



from sys import argv
from time import sleep

from PyQt5 import QtWidgets, QtCore
from PyQt5.QtWidgets import QApplication, QMainWindow, QProxyStyle, QStyle

class CustomStyle(QProxyStyle):
    def styleHint(self, hint, option=None, widget=None, returnData=None):
        if hint == QStyle.SH_SpinBox_KeyPressAutoRepeatRate:
            return 10**6
        elif hint == QStyle.SH_SpinBox_ClickAutoRepeatRate:
            return 10**6
        elif hint == QStyle.SH_SpinBox_ClickAutoRepeatThreshold:
            # You can use only this condition to avoid the auto-repeat,
            # but better safe than sorry ;-)
            return 10**6
        else:
            return super().styleHint(hint, option, widget, returnData)

class SpinBoxTest(QtWidgets.QWidget):
    def __init__(self):
        super(SpinBoxTest, self).__init__()
        self.spinBox = QtWidgets.QSpinBox(self)
        self.spinBox.setGeometry(QtCore.QRect(9, 27, 131, 20))
        self.spinBox.setStyle(CustomStyle())
        self.spinBox.valueChanged.connect(self.spinbox_changed)

    def spinbox_changed(self):
        sleep(0.5)
        print("spinbox has changed, value: " + str(self.spinBox.value()))

if __name__ == '__main__':
    app = QtWidgets.QApplication(argv)
    window = SpinBoxTest()
    window.show()
    app.exec_()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'job_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.12
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_JobDialog(object):
    def setupUi(self, JobDialog):
        JobDialog.setObjectName("JobDialog")
        JobDialog.resize(900, 530)
        self.gridLayout = QtWidgets.QGridLayout(JobDialog)
        self.gridLayout.setObjectName("gridLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 1, 2, 1, 1)
        self.job_list_widget = QtWidgets.QListWidget(JobDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.job_list_widget.sizePolicy().hasHeightForWidth())
        self.job_list_widget.setSizePolicy(sizePolicy)
        self.job_list_widget.setMinimumSize(QtCore.QSize(0, 483))
        self.job_list_widget.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.job_list_widget.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.job_list_widget.setObjectName("job_list_widget")
        self.gridLayout.addWidget(self.job_list_widget, 0, 0, 1, 6)
        self.button_remove_jobs = QtWidgets.QPushButton(JobDialog)
        self.button_remove_jobs.setObjectName("button_remove_jobs")
        self.gridLayout.addWidget(self.button_remove_jobs, 1, 1, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(JobDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 1, 5, 1, 1)
        self.button_add_jobs = QtWidgets.QPushButton(JobDialog)
        self.button_add_jobs.setObjectName("button_add_jobs")
        self.gridLayout.addWidget(self.button_add_jobs, 1, 0, 1, 1)
        self.messageLabel = QtWidgets.QLabel(JobDialog)
        self.messageLabel.setObjectName("messageLabel")
        self.gridLayout.addWidget(self.messageLabel, 1, 3, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem1, 1, 4, 1, 1)

        self.retranslateUi(JobDialog)
        self.buttonBox.accepted.connect(JobDialog.accept)
        self.buttonBox.rejected.connect(JobDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(JobDialog)

    def retranslateUi(self, JobDialog):
        _translate = QtCore.QCoreApplication.translate
        JobDialog.setWindowTitle(_translate("JobDialog", "Manage Job List"))
        self.button_remove_jobs.setToolTip(_translate("JobDialog", "Select one or more items in the list. Press this button to remove them from the list."))
        self.button_remove_jobs.setText(_translate("JobDialog", "Remove selected job(s)"))
        self.buttonBox.setToolTip(_translate("JobDialog", "Press \'OK\' to save changes and exit the job editor, or \'Cancel\' to exit without saving."))
        self.button_add_jobs.setToolTip(_translate("JobDialog", "Select job input. This can be video files and / or directories containing image files for stacking, or single image files for postprocessing.\n"
"Each entry is executed as a job, either in batch mode (automatic) or manually. When the list is complete, confirm with pressing \"OK\"."))
        self.button_add_jobs.setText(_translate("JobDialog", "Add job(s)"))
        self.messageLabel.setText(_translate("JobDialog", "Add / remove videos / image folders for stacking, or images for postprocessing, confirm with \'OK\'."))





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'list_view_widget.ui'
#
# Created by: PyQt5 UI code generator 5.11.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_ListViewWidget(object):
    def setupUi(self, ListViewWidget):
        ListViewWidget.setObjectName("ListViewWidget")
        ListViewWidget.resize(264, 341)
        self.horizontalLayout = QtWidgets.QHBoxLayout(ListViewWidget)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.listWidget = QtWidgets.QListWidget(ListViewWidget)
        self.listWidget.setObjectName("listWidget")
        self.horizontalLayout.addWidget(self.listWidget)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.addButton = QtWidgets.QPushButton(ListViewWidget)
        self.addButton.setObjectName("addButton")
        self.verticalLayout.addWidget(self.addButton)
        self.removeButton = QtWidgets.QPushButton(ListViewWidget)
        self.removeButton.setObjectName("removeButton")
        self.verticalLayout.addWidget(self.removeButton)
        self.horizontalLayout.addLayout(self.verticalLayout)

        self.retranslateUi(ListViewWidget)
        QtCore.QMetaObject.connectSlotsByName(ListViewWidget)

    def retranslateUi(self, ListViewWidget):
        _translate = QtCore.QCoreApplication.translate
        ListViewWidget.setWindowTitle(_translate("ListViewWidget", "Frame"))
        self.addButton.setText(_translate("ListViewWidget", "+"))
        self.removeButton.setText(_translate("ListViewWidget", "-"))

